from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    SaveMixin,
    UpdateMethod,
    UpdateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = [
    "ProjectExternalStatusCheck",
    "ProjectExternalStatusCheckManager",
    "ProjectMergeRequestStatusCheck",
    "ProjectMergeRequestStatusCheckManager",
]


class ProjectExternalStatusCheck(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectExternalStatusCheckManager(
    ListMixin[ProjectExternalStatusCheck],
    CreateMixin[ProjectExternalStatusCheck],
    UpdateMixin[ProjectExternalStatusCheck],
    DeleteMixin[ProjectExternalStatusCheck],
):
    _path = "/projects/{project_id}/external_status_checks"
    _obj_cls = ProjectExternalStatusCheck
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "external_url"),
        optional=("shared_secret", "protected_branch_ids"),
    )
    _update_attrs = RequiredOptional(
        optional=("name", "external_url", "shared_secret", "protected_branch_ids")
    )
    _types = {"protected_branch_ids": ArrayAttribute}


class ProjectMergeRequestStatusCheck(SaveMixin, RESTObject):
    pass


class ProjectMergeRequestStatusCheckManager(ListMixin[ProjectMergeRequestStatusCheck]):
    _path = "/projects/{project_id}/merge_requests/{merge_request_iid}/status_checks"
    _obj_cls = ProjectMergeRequestStatusCheck
    _from_parent_attrs = {"project_id": "project_id", "merge_request_iid": "iid"}
    _update_attrs = RequiredOptional(
        required=("sha", "external_status_check_id", "status")
    )
    _update_method = UpdateMethod.POST



from gitlab.base import RESTObject
from gitlab.mixins import CreateMixin, RetrieveMixin, SaveMixin, UpdateMixin
from gitlab.types import RequiredOptional

from .notes import (  # noqa: F401
    ProjectCommitDiscussionNoteManager,
    ProjectIssueDiscussionNoteManager,
    ProjectMergeRequestDiscussionNoteManager,
    ProjectSnippetDiscussionNoteManager,
)

__all__ = [
    "ProjectCommitDiscussion",
    "ProjectCommitDiscussionManager",
    "ProjectIssueDiscussion",
    "ProjectIssueDiscussionManager",
    "ProjectMergeRequestDiscussion",
    "ProjectMergeRequestDiscussionManager",
    "ProjectSnippetDiscussion",
    "ProjectSnippetDiscussionManager",
]


class ProjectCommitDiscussion(RESTObject):
    notes: ProjectCommitDiscussionNoteManager


class ProjectCommitDiscussionManager(
    RetrieveMixin[ProjectCommitDiscussion], CreateMixin[ProjectCommitDiscussion]
):
    _path = "/projects/{project_id}/repository/commits/{commit_id}/discussions"
    _obj_cls = ProjectCommitDiscussion
    _from_parent_attrs = {"project_id": "project_id", "commit_id": "id"}
    _create_attrs = RequiredOptional(required=("body",), optional=("created_at",))


class ProjectIssueDiscussion(RESTObject):
    notes: ProjectIssueDiscussionNoteManager


class ProjectIssueDiscussionManager(
    RetrieveMixin[ProjectIssueDiscussion], CreateMixin[ProjectIssueDiscussion]
):
    _path = "/projects/{project_id}/issues/{issue_iid}/discussions"
    _obj_cls = ProjectIssueDiscussion
    _from_parent_attrs = {"project_id": "project_id", "issue_iid": "iid"}
    _create_attrs = RequiredOptional(required=("body",), optional=("created_at",))


class ProjectMergeRequestDiscussion(SaveMixin, RESTObject):
    notes: ProjectMergeRequestDiscussionNoteManager


class ProjectMergeRequestDiscussionManager(
    RetrieveMixin[ProjectMergeRequestDiscussion],
    CreateMixin[ProjectMergeRequestDiscussion],
    UpdateMixin[ProjectMergeRequestDiscussion],
):
    _path = "/projects/{project_id}/merge_requests/{mr_iid}/discussions"
    _obj_cls = ProjectMergeRequestDiscussion
    _from_parent_attrs = {"project_id": "project_id", "mr_iid": "iid"}
    _create_attrs = RequiredOptional(
        required=("body",), optional=("created_at", "position")
    )
    _update_attrs = RequiredOptional(required=("resolved",))


class ProjectSnippetDiscussion(RESTObject):
    notes: ProjectSnippetDiscussionNoteManager


class ProjectSnippetDiscussionManager(
    RetrieveMixin[ProjectSnippetDiscussion], CreateMixin[ProjectSnippetDiscussion]
):
    _path = "/projects/{project_id}/snippets/{snippet_id}/discussions"
    _obj_cls = ProjectSnippetDiscussion
    _from_parent_attrs = {"project_id": "project_id", "snippet_id": "id"}
    _create_attrs = RequiredOptional(required=("body",), optional=("created_at",))



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from gitlab import exceptions as exc
from gitlab import types
from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    CRUDMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    SaveMixin,
    UpdateMixin,
)
from gitlab.types import RequiredOptional

from .events import GroupEpicResourceLabelEventManager  # noqa: F401
from .notes import GroupEpicNoteManager  # noqa: F401

__all__ = ["GroupEpic", "GroupEpicManager", "GroupEpicIssue", "GroupEpicIssueManager"]


class GroupEpic(ObjectDeleteMixin, SaveMixin, RESTObject):
    _id_attr = "iid"

    issues: GroupEpicIssueManager
    resourcelabelevents: GroupEpicResourceLabelEventManager
    notes: GroupEpicNoteManager


class GroupEpicManager(CRUDMixin[GroupEpic]):
    _path = "/groups/{group_id}/epics"
    _obj_cls = GroupEpic
    _from_parent_attrs = {"group_id": "id"}
    _list_filters = ("author_id", "labels", "order_by", "sort", "search")
    _create_attrs = RequiredOptional(
        required=("title",),
        optional=("labels", "description", "start_date", "end_date"),
    )
    _update_attrs = RequiredOptional(
        optional=("title", "labels", "description", "start_date", "end_date")
    )
    _types = {"labels": types.CommaSeparatedListAttribute}


class GroupEpicIssue(ObjectDeleteMixin, SaveMixin, RESTObject):
    _id_attr = "epic_issue_id"
    # Define type for 'manager' here So mypy won't complain about
    # 'self.manager.update()' call in the 'save' method.
    manager: GroupEpicIssueManager

    def save(self, **kwargs: Any) -> None:
        """Save the changes made to the object to the server.

        The object is updated to match what the server returns.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raise:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        updated_data = self._get_updated_data()
        # Nothing to update. Server fails if sent an empty dict.
        if not updated_data:
            return

        # call the manager
        obj_id = self.encoded_id
        self.manager.update(obj_id, updated_data, **kwargs)


class GroupEpicIssueManager(
    ListMixin[GroupEpicIssue],
    CreateMixin[GroupEpicIssue],
    UpdateMixin[GroupEpicIssue],
    DeleteMixin[GroupEpicIssue],
):
    _path = "/groups/{group_id}/epics/{epic_iid}/issues"
    _obj_cls = GroupEpicIssue
    _from_parent_attrs = {"group_id": "group_id", "epic_iid": "iid"}
    _create_attrs = RequiredOptional(required=("issue_id",))
    _update_attrs = RequiredOptional(optional=("move_before_id", "move_after_id"))

    @exc.on_http_error(exc.GitlabCreateError)
    def create(
        self, data: dict[str, Any] | None = None, **kwargs: Any
    ) -> GroupEpicIssue:
        """Create a new object.

        Args:
            data: Parameters to send to the server to create the
                         resource
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server cannot perform the request

        Returns:
            A new instance of the manage object class build with
                the data sent by the server
        """
        if TYPE_CHECKING:
            assert data is not None
        self._create_attrs.validate_attrs(data=data)
        path = f"{self.path}/{data.pop('issue_id')}"
        server_data = self.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        # The epic_issue_id attribute doesn't exist when creating the resource,
        # but is used everywhere elese. Let's create it to be consistent client
        # side
        server_data["epic_issue_id"] = server_data["id"]
        return self._obj_cls(self, server_data)



from gitlab.base import RESTObject
from gitlab.mixins import GetWithoutIdMixin, SaveMixin, UpdateMixin
from gitlab.types import RequiredOptional

__all__ = [
    "NotificationSettings",
    "NotificationSettingsManager",
    "GroupNotificationSettings",
    "GroupNotificationSettingsManager",
    "ProjectNotificationSettings",
    "ProjectNotificationSettingsManager",
]


class NotificationSettings(SaveMixin, RESTObject):
    _id_attr = None


class NotificationSettingsManager(
    GetWithoutIdMixin[NotificationSettings], UpdateMixin[NotificationSettings]
):
    _path = "/notification_settings"
    _obj_cls = NotificationSettings

    _update_attrs = RequiredOptional(
        optional=(
            "level",
            "notification_email",
            "new_note",
            "new_issue",
            "reopen_issue",
            "close_issue",
            "reassign_issue",
            "new_merge_request",
            "reopen_merge_request",
            "close_merge_request",
            "reassign_merge_request",
            "merge_merge_request",
        )
    )


class GroupNotificationSettings(NotificationSettings):
    pass


class GroupNotificationSettingsManager(NotificationSettingsManager):
    _path = "/groups/{group_id}/notification_settings"
    _obj_cls = GroupNotificationSettings
    _from_parent_attrs = {"group_id": "id"}


class ProjectNotificationSettings(NotificationSettings):
    pass


class ProjectNotificationSettingsManager(NotificationSettingsManager):
    _path = "/projects/{project_id}/notification_settings"
    _obj_cls = ProjectNotificationSettings
    _from_parent_attrs = {"project_id": "id"}



from gitlab.base import RESTObject
from gitlab.mixins import BadgeRenderMixin, CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = ["GroupBadge", "GroupBadgeManager", "ProjectBadge", "ProjectBadgeManager"]


class GroupBadge(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class GroupBadgeManager(BadgeRenderMixin[GroupBadge], CRUDMixin[GroupBadge]):
    _path = "/groups/{group_id}/badges"
    _obj_cls = GroupBadge
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(required=("link_url", "image_url"))
    _update_attrs = RequiredOptional(optional=("link_url", "image_url"))


class ProjectBadge(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectBadgeManager(BadgeRenderMixin[ProjectBadge], CRUDMixin[ProjectBadge]):
    _path = "/projects/{project_id}/badges"
    _obj_cls = ProjectBadge
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("link_url", "image_url"))
    _update_attrs = RequiredOptional(optional=("link_url", "image_url"))



from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ObjectDeleteMixin,
    ObjectRotateMixin,
    RetrieveMixin,
    RotateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = ["ProjectAccessToken", "ProjectAccessTokenManager"]


class ProjectAccessToken(ObjectDeleteMixin, ObjectRotateMixin, RESTObject):
    pass


class ProjectAccessTokenManager(
    CreateMixin[ProjectAccessToken],
    DeleteMixin[ProjectAccessToken],
    RetrieveMixin[ProjectAccessToken],
    RotateMixin[ProjectAccessToken],
):
    _path = "/projects/{project_id}/access_tokens"
    _obj_cls = ProjectAccessToken
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "scopes"), optional=("access_level", "expires_at")
    )
    _types = {"scopes": ArrayAttribute}



"""
GitLab API:
https://docs.gitlab.com/ee/api/lint.html
"""

from typing import Any

from gitlab.base import RESTObject
from gitlab.cli import register_custom_action
from gitlab.exceptions import GitlabCiLintError
from gitlab.mixins import CreateMixin, GetWithoutIdMixin
from gitlab.types import RequiredOptional

__all__ = ["CiLint", "CiLintManager", "ProjectCiLint", "ProjectCiLintManager"]


class CiLint(RESTObject):
    _id_attr = None


class CiLintManager(CreateMixin[CiLint]):
    _path = "/ci/lint"
    _obj_cls = CiLint
    _create_attrs = RequiredOptional(
        required=("content",), optional=("include_merged_yaml", "include_jobs")
    )

    @register_custom_action(
        cls_names="CiLintManager",
        required=("content",),
        optional=("include_merged_yaml", "include_jobs"),
    )
    def validate(self, *args: Any, **kwargs: Any) -> None:
        """Raise an error if the CI Lint results are not valid.

        This is a custom python-gitlab method to wrap lint endpoints."""
        result = self.create(*args, **kwargs)

        if result.status != "valid":
            message = ",\n".join(result.errors)
            raise GitlabCiLintError(message)


class ProjectCiLint(RESTObject):
    _id_attr = None


class ProjectCiLintManager(
    GetWithoutIdMixin[ProjectCiLint], CreateMixin[ProjectCiLint]
):
    _path = "/projects/{project_id}/ci/lint"
    _obj_cls = ProjectCiLint
    _from_parent_attrs = {"project_id": "id"}
    _optional_get_attrs = (
        "content_ref",
        "dry_run",
        "dry_run_ref",
        "include_jobs",
        "ref",
    )

    _create_attrs = RequiredOptional(
        required=("content",), optional=("dry_run", "include_jobs", "ref")
    )

    @register_custom_action(
        cls_names="ProjectCiLintManager",
        required=("content",),
        optional=("dry_run", "include_jobs", "ref"),
    )
    def validate(self, *args: Any, **kwargs: Any) -> None:
        """Raise an error if the Project CI Lint results are not valid.

        This is a custom python-gitlab method to wrap lint endpoints."""
        result = self.create(*args, **kwargs)

        if not result.valid:
            message = ",\n".join(result.errors)
            raise GitlabCiLintError(message)



from __future__ import annotations

from typing import Any

from gitlab import cli
from gitlab import exceptions as exc
from gitlab import types
from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    CRUDMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    SaveMixin,
)
from gitlab.types import RequiredOptional

__all__ = [
    "RunnerJob",
    "RunnerJobManager",
    "Runner",
    "RunnerManager",
    "RunnerAll",
    "RunnerAllManager",
    "GroupRunner",
    "GroupRunnerManager",
    "ProjectRunner",
    "ProjectRunnerManager",
]


class RunnerJob(RESTObject):
    pass


class RunnerJobManager(ListMixin[RunnerJob]):
    _path = "/runners/{runner_id}/jobs"
    _obj_cls = RunnerJob
    _from_parent_attrs = {"runner_id": "id"}
    _list_filters = ("status", "order_by", "sort")


class Runner(SaveMixin, ObjectDeleteMixin, RESTObject):
    jobs: RunnerJobManager
    _repr_attr = "description"


class RunnerManager(CRUDMixin[Runner]):
    _path = "/runners"
    _obj_cls = Runner
    _create_attrs = RequiredOptional(
        required=("token",),
        optional=(
            "description",
            "info",
            "active",
            "locked",
            "run_untagged",
            "tag_list",
            "access_level",
            "maximum_timeout",
        ),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "description",
            "active",
            "tag_list",
            "run_untagged",
            "locked",
            "access_level",
            "maximum_timeout",
        )
    )
    _list_filters = ("scope", "type", "status", "paused", "tag_list")
    _types = {"tag_list": types.CommaSeparatedListAttribute}

    @cli.register_custom_action(cls_names="RunnerManager", optional=("scope",))
    @exc.on_http_error(exc.GitlabListError)
    def all(self, scope: str | None = None, **kwargs: Any) -> list[Runner]:
        """List all the runners.

        Args:
            scope: The scope of runners to show, one of: specific,
                shared, active, paused, online
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            iterator: If set to True and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the server failed to perform the request

        Returns:
            A list of runners matching the scope.
        """
        path = "/runners/all"
        query_data = {}
        if scope is not None:
            query_data["scope"] = scope
        obj = self.gitlab.http_list(path, query_data, **kwargs)
        return [self._obj_cls(self, item) for item in obj]

    @cli.register_custom_action(cls_names="RunnerManager", required=("token",))
    @exc.on_http_error(exc.GitlabVerifyError)
    def verify(self, token: str, **kwargs: Any) -> None:
        """Validates authentication credentials for a registered Runner.

        Args:
            token: The runner's authentication token
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabVerifyError: If the server failed to verify the token
        """
        path = "/runners/verify"
        post_data = {"token": token}
        self.gitlab.http_post(path, post_data=post_data, **kwargs)


class RunnerAll(RESTObject):
    _repr_attr = "description"


class RunnerAllManager(ListMixin[RunnerAll]):
    _path = "/runners/all"
    _obj_cls = RunnerAll
    _list_filters = ("scope", "type", "status", "paused", "tag_list")
    _types = {"tag_list": types.CommaSeparatedListAttribute}


class GroupRunner(RESTObject):
    pass


class GroupRunnerManager(ListMixin[GroupRunner]):
    _path = "/groups/{group_id}/runners"
    _obj_cls = GroupRunner
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(required=("runner_id",))
    _list_filters = ("scope", "tag_list")
    _types = {"tag_list": types.CommaSeparatedListAttribute}


class ProjectRunner(ObjectDeleteMixin, RESTObject):
    pass


class ProjectRunnerManager(
    CreateMixin[ProjectRunner], DeleteMixin[ProjectRunner], ListMixin[ProjectRunner]
):
    _path = "/projects/{project_id}/runners"
    _obj_cls = ProjectRunner
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("runner_id",))
    _list_filters = ("scope", "tag_list")
    _types = {"tag_list": types.CommaSeparatedListAttribute}



"""
GitLab API:
https://docs.gitlab.com/ee/api/deployments.html
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from gitlab import cli
from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import CreateMixin, RetrieveMixin, SaveMixin, UpdateMixin
from gitlab.types import RequiredOptional

from .merge_requests import ProjectDeploymentMergeRequestManager  # noqa: F401

__all__ = ["ProjectDeployment", "ProjectDeploymentManager"]


class ProjectDeployment(SaveMixin, RESTObject):
    mergerequests: ProjectDeploymentMergeRequestManager

    @cli.register_custom_action(
        cls_names="ProjectDeployment",
        required=("status",),
        optional=("comment", "represented_as"),
    )
    @exc.on_http_error(exc.GitlabDeploymentApprovalError)
    def approval(
        self,
        status: str,
        comment: str | None = None,
        represented_as: str | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Approve or reject a blocked deployment.

        Args:
            status: Either "approved" or "rejected"
            comment: A comment to go with the approval
            represented_as: The name of the User/Group/Role to use for the
                            approval, when the user belongs to multiple
                            approval rules.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMRApprovalError: If the approval failed

        Returns:
           A dict containing the result.

        https://docs.gitlab.com/ee/api/deployments.html#approve-or-reject-a-blocked-deployment
        """
        path = f"{self.manager.path}/{self.encoded_id}/approval"
        data = {"status": status}
        if comment is not None:
            data["comment"] = comment
        if represented_as is not None:
            data["represented_as"] = represented_as

        server_data = self.manager.gitlab.http_post(path, post_data=data, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        return server_data


class ProjectDeploymentManager(
    RetrieveMixin[ProjectDeployment],
    CreateMixin[ProjectDeployment],
    UpdateMixin[ProjectDeployment],
):
    _path = "/projects/{project_id}/deployments"
    _obj_cls = ProjectDeployment
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = (
        "order_by",
        "sort",
        "updated_after",
        "updated_before",
        "environment",
        "status",
    )
    _create_attrs = RequiredOptional(
        required=("sha", "ref", "tag", "status", "environment")
    )



"""
GitLab API:
https://docs.gitlab.com/ee/api/merge_requests.html
https://docs.gitlab.com/ee/api/merge_request_approvals.html
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

import requests

import gitlab
from gitlab import cli
from gitlab import exceptions as exc
from gitlab import types
from gitlab.base import RESTObject, RESTObjectList
from gitlab.mixins import (
    CRUDMixin,
    ListMixin,
    ObjectDeleteMixin,
    ParticipantsMixin,
    RetrieveMixin,
    SaveMixin,
    SubscribableMixin,
    TimeTrackingMixin,
    TodoMixin,
)
from gitlab.types import RequiredOptional

from .award_emojis import ProjectMergeRequestAwardEmojiManager  # noqa: F401
from .commits import ProjectCommit, ProjectCommitManager
from .discussions import ProjectMergeRequestDiscussionManager  # noqa: F401
from .draft_notes import ProjectMergeRequestDraftNoteManager
from .events import (  # noqa: F401
    ProjectMergeRequestResourceLabelEventManager,
    ProjectMergeRequestResourceMilestoneEventManager,
    ProjectMergeRequestResourceStateEventManager,
)
from .issues import ProjectIssue, ProjectIssueManager
from .merge_request_approvals import (  # noqa: F401
    ProjectMergeRequestApprovalManager,
    ProjectMergeRequestApprovalRuleManager,
    ProjectMergeRequestApprovalStateManager,
)
from .notes import ProjectMergeRequestNoteManager  # noqa: F401
from .pipelines import ProjectMergeRequestPipelineManager  # noqa: F401
from .reviewers import ProjectMergeRequestReviewerDetailManager
from .status_checks import ProjectMergeRequestStatusCheckManager

__all__ = [
    "MergeRequest",
    "MergeRequestManager",
    "GroupMergeRequest",
    "GroupMergeRequestManager",
    "ProjectMergeRequest",
    "ProjectMergeRequestManager",
    "ProjectDeploymentMergeRequest",
    "ProjectDeploymentMergeRequestManager",
    "ProjectMergeRequestDiff",
    "ProjectMergeRequestDiffManager",
]


class MergeRequest(RESTObject):
    pass


class MergeRequestManager(ListMixin[MergeRequest]):
    _path = "/merge_requests"
    _obj_cls = MergeRequest
    _list_filters = (
        "state",
        "order_by",
        "sort",
        "milestone",
        "view",
        "labels",
        "with_labels_details",
        "with_merge_status_recheck",
        "created_after",
        "created_before",
        "updated_after",
        "updated_before",
        "scope",
        "author_id",
        "author_username",
        "assignee_id",
        "approver_ids",
        "approved_by_ids",
        "reviewer_id",
        "reviewer_username",
        "my_reaction_emoji",
        "source_branch",
        "target_branch",
        "search",
        "in",
        "wip",
        "not",
        "environment",
        "deployed_before",
        "deployed_after",
    )
    _types = {
        "approver_ids": types.ArrayAttribute,
        "approved_by_ids": types.ArrayAttribute,
        "in": types.CommaSeparatedListAttribute,
        "labels": types.CommaSeparatedListAttribute,
    }


class GroupMergeRequest(RESTObject):
    pass


class GroupMergeRequestManager(ListMixin[GroupMergeRequest]):
    _path = "/groups/{group_id}/merge_requests"
    _obj_cls = GroupMergeRequest
    _from_parent_attrs = {"group_id": "id"}
    _list_filters = (
        "state",
        "order_by",
        "sort",
        "milestone",
        "view",
        "labels",
        "created_after",
        "created_before",
        "updated_after",
        "updated_before",
        "scope",
        "author_id",
        "assignee_id",
        "approver_ids",
        "approved_by_ids",
        "my_reaction_emoji",
        "source_branch",
        "target_branch",
        "search",
        "wip",
    )
    _types = {
        "approver_ids": types.ArrayAttribute,
        "approved_by_ids": types.ArrayAttribute,
        "labels": types.CommaSeparatedListAttribute,
    }


class ProjectMergeRequest(
    SubscribableMixin,
    TodoMixin,
    TimeTrackingMixin,
    ParticipantsMixin,
    SaveMixin,
    ObjectDeleteMixin,
    RESTObject,
):
    _id_attr = "iid"

    approval_rules: ProjectMergeRequestApprovalRuleManager
    approval_state: ProjectMergeRequestApprovalStateManager
    approvals: ProjectMergeRequestApprovalManager
    awardemojis: ProjectMergeRequestAwardEmojiManager
    diffs: ProjectMergeRequestDiffManager
    discussions: ProjectMergeRequestDiscussionManager
    draft_notes: ProjectMergeRequestDraftNoteManager
    notes: ProjectMergeRequestNoteManager
    pipelines: ProjectMergeRequestPipelineManager
    resourcelabelevents: ProjectMergeRequestResourceLabelEventManager
    resourcemilestoneevents: ProjectMergeRequestResourceMilestoneEventManager
    resourcestateevents: ProjectMergeRequestResourceStateEventManager
    reviewer_details: ProjectMergeRequestReviewerDetailManager
    status_checks: ProjectMergeRequestStatusCheckManager

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabMROnBuildSuccessError)
    def cancel_merge_when_pipeline_succeeds(self, **kwargs: Any) -> dict[str, str]:
        """Cancel merge when the pipeline succeeds.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMROnBuildSuccessError: If the server could not handle the
                request

        Returns:
            dict of the parsed json returned by the server
        """

        path = (
            f"{self.manager.path}/{self.encoded_id}/cancel_merge_when_pipeline_succeeds"
        )
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        # 2022-10-30: The docs at
        # https://docs.gitlab.com/ee/api/merge_requests.html#cancel-merge-when-pipeline-succeeds
        # are incorrect in that the return value is actually just:
        #   {'status': 'success'}  for a successful cancel.
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        return server_data

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabListError)
    def related_issues(self, **kwargs: Any) -> RESTObjectList[ProjectIssue]:
        """List issues related to this merge request."

        Args:
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the list could not be retrieved

        Returns:
            List of issues
        """

        path = f"{self.manager.path}/{self.encoded_id}/related_issues"
        data_list = self.manager.gitlab.http_list(path, iterator=True, **kwargs)

        if TYPE_CHECKING:
            assert isinstance(data_list, gitlab.GitlabList)

        manager = ProjectIssueManager(self.manager.gitlab, parent=self.manager._parent)

        return RESTObjectList(manager, ProjectIssue, data_list)

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabListError)
    def closes_issues(self, **kwargs: Any) -> RESTObjectList[ProjectIssue]:
        """List issues that will close on merge."

        Args:
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the list could not be retrieved

        Returns:
            List of issues
        """
        path = f"{self.manager.path}/{self.encoded_id}/closes_issues"
        data_list = self.manager.gitlab.http_list(path, iterator=True, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(data_list, gitlab.GitlabList)
        manager = ProjectIssueManager(self.manager.gitlab, parent=self.manager._parent)
        return RESTObjectList(manager, ProjectIssue, data_list)

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabListError)
    def commits(self, **kwargs: Any) -> RESTObjectList[ProjectCommit]:
        """List the merge request commits.

        Args:
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the list could not be retrieved

        Returns:
            The list of commits
        """

        path = f"{self.manager.path}/{self.encoded_id}/commits"
        data_list = self.manager.gitlab.http_list(path, iterator=True, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(data_list, gitlab.GitlabList)
        manager = ProjectCommitManager(self.manager.gitlab, parent=self.manager._parent)
        return RESTObjectList(manager, ProjectCommit, data_list)

    @cli.register_custom_action(
        cls_names="ProjectMergeRequest", optional=("access_raw_diffs",)
    )
    @exc.on_http_error(exc.GitlabListError)
    def changes(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """List the merge request changes.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the list could not be retrieved

        Returns:
            List of changes
        """
        path = f"{self.manager.path}/{self.encoded_id}/changes"
        return self.manager.gitlab.http_get(path, **kwargs)

    @cli.register_custom_action(cls_names="ProjectMergeRequest", optional=("sha",))
    @exc.on_http_error(exc.GitlabMRApprovalError)
    def approve(self, sha: str | None = None, **kwargs: Any) -> dict[str, Any]:
        """Approve the merge request.

        Args:
            sha: Head SHA of MR
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMRApprovalError: If the approval failed

        Returns:
           A dict containing the result.

        https://docs.gitlab.com/ee/api/merge_request_approvals.html#approve-merge-request
        """
        path = f"{self.manager.path}/{self.encoded_id}/approve"
        data = {}
        if sha:
            data["sha"] = sha

        server_data = self.manager.gitlab.http_post(path, post_data=data, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        self._update_attrs(server_data)
        return server_data

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabMRApprovalError)
    def unapprove(self, **kwargs: Any) -> None:
        """Unapprove the merge request.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMRApprovalError: If the unapproval failed

        https://docs.gitlab.com/ee/api/merge_request_approvals.html#unapprove-merge-request
        """
        path = f"{self.manager.path}/{self.encoded_id}/unapprove"
        data: dict[str, Any] = {}

        server_data = self.manager.gitlab.http_post(path, post_data=data, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        self._update_attrs(server_data)

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabMRRebaseError)
    def rebase(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Attempt to rebase the source branch onto the target branch

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMRRebaseError: If rebasing failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/rebase"
        data: dict[str, Any] = {}
        return self.manager.gitlab.http_put(path, post_data=data, **kwargs)

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabMRResetApprovalError)
    def reset_approvals(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Clear all approvals of the merge request.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMRResetApprovalError: If reset approval failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/reset_approvals"
        data: dict[str, Any] = {}
        return self.manager.gitlab.http_put(path, post_data=data, **kwargs)

    @cli.register_custom_action(cls_names="ProjectMergeRequest")
    @exc.on_http_error(exc.GitlabGetError)
    def merge_ref(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Attempt to merge changes between source and target branches into
            `refs/merge-requests/:iid/merge`.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabGetError: If cannot be merged
        """
        path = f"{self.manager.path}/{self.encoded_id}/merge_ref"
        return self.manager.gitlab.http_get(path, **kwargs)

    @cli.register_custom_action(
        cls_names="ProjectMergeRequest",
        optional=(
            "merge_commit_message",
            "should_remove_source_branch",
            "merge_when_pipeline_succeeds",
        ),
    )
    @exc.on_http_error(exc.GitlabMRClosedError)
    def merge(
        self,
        merge_commit_message: str | None = None,
        should_remove_source_branch: bool | None = None,
        merge_when_pipeline_succeeds: bool | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Accept the merge request.

        Args:
            merge_commit_message: Commit message
            should_remove_source_branch: If True, removes the source
                                                branch
            merge_when_pipeline_succeeds: Wait for the build to succeed,
                                                 then merge
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMRClosedError: If the merge failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/merge"
        data: dict[str, Any] = {}
        if merge_commit_message:
            data["merge_commit_message"] = merge_commit_message
        if should_remove_source_branch is not None:
            data["should_remove_source_branch"] = should_remove_source_branch
        if merge_when_pipeline_succeeds is not None:
            data["merge_when_pipeline_succeeds"] = merge_when_pipeline_succeeds

        server_data = self.manager.gitlab.http_put(path, post_data=data, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        self._update_attrs(server_data)
        return server_data


class ProjectMergeRequestManager(CRUDMixin[ProjectMergeRequest]):
    _path = "/projects/{project_id}/merge_requests"
    _obj_cls = ProjectMergeRequest
    _from_parent_attrs = {"project_id": "id"}
    _optional_get_attrs = (
        "render_html",
        "include_diverged_commits_count",
        "include_rebase_in_progress",
    )
    _create_attrs = RequiredOptional(
        required=("source_branch", "target_branch", "title"),
        optional=(
            "allow_collaboration",
            "allow_maintainer_to_push",
            "approvals_before_merge",
            "assignee_id",
            "assignee_ids",
            "description",
            "labels",
            "milestone_id",
            "remove_source_branch",
            "reviewer_ids",
            "squash",
            "target_project_id",
        ),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "target_branch",
            "assignee_id",
            "title",
            "description",
            "state_event",
            "labels",
            "milestone_id",
            "remove_source_branch",
            "discussion_locked",
            "allow_maintainer_to_push",
            "squash",
            "reviewer_ids",
        )
    )
    _list_filters = (
        "state",
        "order_by",
        "sort",
        "milestone",
        "view",
        "labels",
        "created_after",
        "created_before",
        "updated_after",
        "updated_before",
        "scope",
        "iids",
        "author_id",
        "assignee_id",
        "approver_ids",
        "approved_by_ids",
        "my_reaction_emoji",
        "source_branch",
        "target_branch",
        "search",
        "wip",
    )
    _types = {
        "approver_ids": types.ArrayAttribute,
        "approved_by_ids": types.ArrayAttribute,
        "iids": types.ArrayAttribute,
        "labels": types.CommaSeparatedListAttribute,
    }


class ProjectDeploymentMergeRequest(MergeRequest):
    pass


class ProjectDeploymentMergeRequestManager(MergeRequestManager):
    _path = "/projects/{project_id}/deployments/{deployment_id}/merge_requests"
    _obj_cls = ProjectDeploymentMergeRequest
    _from_parent_attrs = {"deployment_id": "id", "project_id": "project_id"}


class ProjectMergeRequestDiff(RESTObject):
    pass


class ProjectMergeRequestDiffManager(RetrieveMixin[ProjectMergeRequestDiff]):
    _path = "/projects/{project_id}/merge_requests/{mr_iid}/versions"
    _obj_cls = ProjectMergeRequestDiff
    _from_parent_attrs = {"project_id": "project_id", "mr_iid": "iid"}



from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = ["BroadcastMessage", "BroadcastMessageManager"]


class BroadcastMessage(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class BroadcastMessageManager(CRUDMixin[BroadcastMessage]):
    _path = "/broadcast_messages"
    _obj_cls = BroadcastMessage

    _create_attrs = RequiredOptional(
        required=("message",),
        optional=("starts_at", "ends_at", "color", "font", "target_access_levels"),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "message",
            "starts_at",
            "ends_at",
            "color",
            "font",
            "target_access_levels",
        )
    )
    _types = {"target_access_levels": ArrayAttribute}



from gitlab.base import RESTObject
from gitlab.mixins import NoUpdateMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = ["ProjectClusterAgent", "ProjectClusterAgentManager"]


class ProjectClusterAgent(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "name"


class ProjectClusterAgentManager(NoUpdateMixin[ProjectClusterAgent]):
    _path = "/projects/{project_id}/cluster_agents"
    _obj_cls = ProjectClusterAgent
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("name",))



from __future__ import annotations

from typing import Any

from gitlab.base import RESTObject, TObjCls
from gitlab.exceptions import GitlabInvitationError
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import ArrayAttribute, CommaSeparatedListAttribute, RequiredOptional

__all__ = [
    "ProjectInvitation",
    "ProjectInvitationManager",
    "GroupInvitation",
    "GroupInvitationManager",
]


class InvitationMixin(CRUDMixin[TObjCls]):
    # pylint: disable=abstract-method
    def create(self, data: dict[str, Any] | None = None, **kwargs: Any) -> TObjCls:
        invitation = super().create(data, **kwargs)

        if invitation.status == "error":
            raise GitlabInvitationError(invitation.message)

        return invitation


class ProjectInvitation(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "email"


class ProjectInvitationManager(InvitationMixin[ProjectInvitation]):
    _path = "/projects/{project_id}/invitations"
    _obj_cls = ProjectInvitation
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("access_level",),
        optional=(
            "expires_at",
            "invite_source",
            "tasks_to_be_done",
            "tasks_project_id",
        ),
        exclusive=("email", "user_id"),
    )
    _update_attrs = RequiredOptional(optional=("access_level", "expires_at"))
    _list_filters = ("query",)
    _types = {
        "email": CommaSeparatedListAttribute,
        "user_id": CommaSeparatedListAttribute,
        "tasks_to_be_done": ArrayAttribute,
    }


class GroupInvitation(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "email"


class GroupInvitationManager(InvitationMixin[GroupInvitation]):
    _path = "/groups/{group_id}/invitations"
    _obj_cls = GroupInvitation
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("access_level",),
        optional=(
            "expires_at",
            "invite_source",
            "tasks_to_be_done",
            "tasks_project_id",
        ),
        exclusive=("email", "user_id"),
    )
    _update_attrs = RequiredOptional(optional=("access_level", "expires_at"))
    _list_filters = ("query",)
    _types = {
        "email": CommaSeparatedListAttribute,
        "user_id": CommaSeparatedListAttribute,
        "tasks_to_be_done": ArrayAttribute,
    }



"""
GitLab API:
https://docs.gitlab.com/ee/api/instance_level_ci_variables.html
https://docs.gitlab.com/ee/api/project_level_variables.html
https://docs.gitlab.com/ee/api/group_level_variables.html
"""

from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "Variable",
    "VariableManager",
    "GroupVariable",
    "GroupVariableManager",
    "ProjectVariable",
    "ProjectVariableManager",
]


class Variable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class VariableManager(CRUDMixin[Variable]):
    _path = "/admin/ci/variables"
    _obj_cls = Variable
    _create_attrs = RequiredOptional(
        required=("key", "value"),
        optional=("description", "masked", "protected", "raw", "variable_type"),
    )
    _update_attrs = RequiredOptional(
        required=("key",),
        optional=(
            "description",
            "masked",
            "protected",
            "raw",
            "value",
            "variable_type",
        ),
    )


class GroupVariable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class GroupVariableManager(CRUDMixin[GroupVariable]):
    _path = "/groups/{group_id}/variables"
    _obj_cls = GroupVariable
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("key", "value"),
        optional=(
            "description",
            "environment_scope",
            "masked",
            "masked_and_hidden",
            "protected",
            "raw",
            "variable_type",
        ),
    )
    _update_attrs = RequiredOptional(
        required=("key",),
        optional=(
            "description",
            "environment_scope",
            "masked",
            "protected",
            "raw",
            "value",
            "variable_type",
        ),
    )


class ProjectVariable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class ProjectVariableManager(CRUDMixin[ProjectVariable]):
    _path = "/projects/{project_id}/variables"
    _obj_cls = ProjectVariable
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("key", "value"),
        optional=(
            "description",
            "environment_scope",
            "masked",
            "masked_and_hidden",
            "protected",
            "raw",
            "variable_type",
        ),
    )
    _update_attrs = RequiredOptional(
        required=("key",),
        optional=(
            "description",
            "environment_scope",
            "masked",
            "protected",
            "raw",
            "value",
            "variable_type",
        ),
    )



from __future__ import annotations

from typing import Any, TYPE_CHECKING

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    CRUDMixin,
    DeleteMixin,
    GetWithoutIdMixin,
    ListMixin,
    ObjectDeleteMixin,
    RefreshMixin,
    RetrieveMixin,
    SaveMixin,
    UpdateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = [
    "ProjectMergeRequestPipeline",
    "ProjectMergeRequestPipelineManager",
    "ProjectPipeline",
    "ProjectPipelineManager",
    "ProjectPipelineJob",
    "ProjectPipelineJobManager",
    "ProjectPipelineBridge",
    "ProjectPipelineBridgeManager",
    "ProjectPipelineVariable",
    "ProjectPipelineVariableManager",
    "ProjectPipelineScheduleVariable",
    "ProjectPipelineScheduleVariableManager",
    "ProjectPipelineSchedulePipeline",
    "ProjectPipelineSchedulePipelineManager",
    "ProjectPipelineSchedule",
    "ProjectPipelineScheduleManager",
    "ProjectPipelineTestReport",
    "ProjectPipelineTestReportManager",
    "ProjectPipelineTestReportSummary",
    "ProjectPipelineTestReportSummaryManager",
]


class ProjectMergeRequestPipeline(RESTObject):
    pass


class ProjectMergeRequestPipelineManager(
    CreateMixin[ProjectMergeRequestPipeline], ListMixin[ProjectMergeRequestPipeline]
):
    _path = "/projects/{project_id}/merge_requests/{mr_iid}/pipelines"
    _obj_cls = ProjectMergeRequestPipeline
    _from_parent_attrs = {"project_id": "project_id", "mr_iid": "iid"}


class ProjectPipeline(RefreshMixin, ObjectDeleteMixin, RESTObject):
    bridges: ProjectPipelineBridgeManager
    jobs: ProjectPipelineJobManager
    test_report: ProjectPipelineTestReportManager
    test_report_summary: ProjectPipelineTestReportSummaryManager
    variables: ProjectPipelineVariableManager

    @cli.register_custom_action(cls_names="ProjectPipeline")
    @exc.on_http_error(exc.GitlabPipelineCancelError)
    def cancel(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Cancel the job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabPipelineCancelError: If the request failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/cancel"
        return self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="ProjectPipeline")
    @exc.on_http_error(exc.GitlabPipelineRetryError)
    def retry(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Retry the job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabPipelineRetryError: If the request failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/retry"
        return self.manager.gitlab.http_post(path, **kwargs)


class ProjectPipelineManager(
    RetrieveMixin[ProjectPipeline],
    CreateMixin[ProjectPipeline],
    DeleteMixin[ProjectPipeline],
):
    _path = "/projects/{project_id}/pipelines"
    _obj_cls = ProjectPipeline
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = (
        "scope",
        "status",
        "source",
        "ref",
        "sha",
        "yaml_errors",
        "name",
        "username",
        "order_by",
        "sort",
    )
    _create_attrs = RequiredOptional(required=("ref",))

    def create(
        self, data: dict[str, Any] | None = None, **kwargs: Any
    ) -> ProjectPipeline:
        """Creates a new object.

        Args:
            data: Parameters to send to the server to create the
                         resource
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server cannot perform the request

        Returns:
            A new instance of the managed object class build with
                the data sent by the server
        """
        path = self.path[:-1]  # drop the 's'
        return super().create(data, path=path, **kwargs)

    def latest(self, ref: str | None = None, lazy: bool = False) -> ProjectPipeline:
        """Get the latest pipeline for the most recent commit
                            on a specific ref in a project

        Args:
            ref: The branch or tag to check for the latest pipeline.
                            Defaults to the default branch when not specified.
        Returns:
            A Pipeline instance
        """
        data = {}
        if ref:
            data = {"ref": ref}
        server_data = self.gitlab.http_get(self.path + "/latest", query_data=data)
        if TYPE_CHECKING:
            assert not isinstance(server_data, requests.Response)
        return self._obj_cls(self, server_data, lazy=lazy)


class ProjectPipelineJob(RESTObject):
    pass


class ProjectPipelineJobManager(ListMixin[ProjectPipelineJob]):
    _path = "/projects/{project_id}/pipelines/{pipeline_id}/jobs"
    _obj_cls = ProjectPipelineJob
    _from_parent_attrs = {"project_id": "project_id", "pipeline_id": "id"}
    _list_filters = ("scope", "include_retried")
    _types = {"scope": ArrayAttribute}


class ProjectPipelineBridge(RESTObject):
    pass


class ProjectPipelineBridgeManager(ListMixin[ProjectPipelineBridge]):
    _path = "/projects/{project_id}/pipelines/{pipeline_id}/bridges"
    _obj_cls = ProjectPipelineBridge
    _from_parent_attrs = {"project_id": "project_id", "pipeline_id": "id"}
    _list_filters = ("scope",)


class ProjectPipelineVariable(RESTObject):
    _id_attr = "key"


class ProjectPipelineVariableManager(ListMixin[ProjectPipelineVariable]):
    _path = "/projects/{project_id}/pipelines/{pipeline_id}/variables"
    _obj_cls = ProjectPipelineVariable
    _from_parent_attrs = {"project_id": "project_id", "pipeline_id": "id"}


class ProjectPipelineScheduleVariable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class ProjectPipelineScheduleVariableManager(
    CreateMixin[ProjectPipelineScheduleVariable],
    UpdateMixin[ProjectPipelineScheduleVariable],
    DeleteMixin[ProjectPipelineScheduleVariable],
):
    _path = "/projects/{project_id}/pipeline_schedules/{pipeline_schedule_id}/variables"
    _obj_cls = ProjectPipelineScheduleVariable
    _from_parent_attrs = {"project_id": "project_id", "pipeline_schedule_id": "id"}
    _create_attrs = RequiredOptional(required=("key", "value"))
    _update_attrs = RequiredOptional(required=("key", "value"))


class ProjectPipelineSchedulePipeline(RESTObject):
    pass


class ProjectPipelineSchedulePipelineManager(
    ListMixin[ProjectPipelineSchedulePipeline]
):
    _path = "/projects/{project_id}/pipeline_schedules/{pipeline_schedule_id}/pipelines"
    _obj_cls = ProjectPipelineSchedulePipeline
    _from_parent_attrs = {"project_id": "project_id", "pipeline_schedule_id": "id"}


class ProjectPipelineSchedule(SaveMixin, ObjectDeleteMixin, RESTObject):
    variables: ProjectPipelineScheduleVariableManager
    pipelines: ProjectPipelineSchedulePipelineManager

    @cli.register_custom_action(cls_names="ProjectPipelineSchedule")
    @exc.on_http_error(exc.GitlabOwnershipError)
    def take_ownership(self, **kwargs: Any) -> None:
        """Update the owner of a pipeline schedule.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabOwnershipError: If the request failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/take_ownership"
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        self._update_attrs(server_data)

    @cli.register_custom_action(cls_names="ProjectPipelineSchedule")
    @exc.on_http_error(exc.GitlabPipelinePlayError)
    def play(self, **kwargs: Any) -> dict[str, Any]:
        """Trigger a new scheduled pipeline, which runs immediately.
        The next scheduled run of this pipeline is not affected.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabPipelinePlayError: If the request failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/play"
        server_data = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        self._update_attrs(server_data)
        return server_data


class ProjectPipelineScheduleManager(CRUDMixin[ProjectPipelineSchedule]):
    _path = "/projects/{project_id}/pipeline_schedules"
    _obj_cls = ProjectPipelineSchedule
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("description", "ref", "cron"), optional=("cron_timezone", "active")
    )
    _update_attrs = RequiredOptional(
        optional=("description", "ref", "cron", "cron_timezone", "active")
    )


class ProjectPipelineTestReport(RESTObject):
    _id_attr = None


class ProjectPipelineTestReportManager(GetWithoutIdMixin[ProjectPipelineTestReport]):
    _path = "/projects/{project_id}/pipelines/{pipeline_id}/test_report"
    _obj_cls = ProjectPipelineTestReport
    _from_parent_attrs = {"project_id": "project_id", "pipeline_id": "id"}


class ProjectPipelineTestReportSummary(RESTObject):
    _id_attr = None


class ProjectPipelineTestReportSummaryManager(
    GetWithoutIdMixin[ProjectPipelineTestReportSummary]
):
    _path = "/projects/{project_id}/pipelines/{pipeline_id}/test_report_summary"
    _obj_cls = ProjectPipelineTestReportSummary
    _from_parent_attrs = {"project_id": "project_id", "pipeline_id": "id"}



"""
GitLab API:
https://docs.gitlab.com/ee/api/packages.html
https://docs.gitlab.com/ee/user/packages/generic_packages/
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, BinaryIO, Callable, Iterator, Literal, overload, TYPE_CHECKING

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab import utils
from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import DeleteMixin, GetMixin, ListMixin, ObjectDeleteMixin

__all__ = [
    "GenericPackage",
    "GenericPackageManager",
    "GroupPackage",
    "GroupPackageManager",
    "ProjectPackage",
    "ProjectPackageManager",
    "ProjectPackageFile",
    "ProjectPackageFileManager",
    "ProjectPackagePipeline",
    "ProjectPackagePipelineManager",
]


class GenericPackage(RESTObject):
    _id_attr = "package_name"


class GenericPackageManager(RESTManager[GenericPackage]):
    _path = "/projects/{project_id}/packages/generic"
    _obj_cls = GenericPackage
    _from_parent_attrs = {"project_id": "id"}

    @cli.register_custom_action(
        cls_names="GenericPackageManager",
        required=("package_name", "package_version", "file_name", "path"),
    )
    @exc.on_http_error(exc.GitlabUploadError)
    def upload(
        self,
        package_name: str,
        package_version: str,
        file_name: str,
        path: str | Path | None = None,
        select: str | None = None,
        data: bytes | BinaryIO | None = None,
        **kwargs: Any,
    ) -> GenericPackage:
        """Upload a file as a generic package.

        Args:
            package_name: The package name. Must follow generic package
                                name regex rules
            package_version: The package version. Must follow semantic
                                version regex rules
            file_name: The name of the file as uploaded in the registry
            path: The path to a local file to upload
            select: GitLab API accepts a value of 'package_file'

        Raises:
            GitlabConnectionError: If the server cannot be reached
            GitlabUploadError: If the file upload fails
            GitlabUploadError: If ``path`` cannot be read
            GitlabUploadError: If both ``path`` and ``data`` are passed

        Returns:
            An object storing the metadata of the uploaded package.

        https://docs.gitlab.com/ee/user/packages/generic_packages/
        """

        if path is None and data is None:
            raise exc.GitlabUploadError("No file contents or path specified")

        if path is not None and data is not None:
            raise exc.GitlabUploadError("File contents and file path specified")

        file_data: bytes | BinaryIO | None = data

        if not file_data:
            if TYPE_CHECKING:
                assert path is not None

            try:
                with open(path, "rb") as f:
                    file_data = f.read()
            except OSError as e:
                raise exc.GitlabUploadError(
                    f"Failed to read package file {path}"
                ) from e

        url = f"{self._computed_path}/{package_name}/{package_version}/{file_name}"
        query_data = {} if select is None else {"select": select}
        server_data = self.gitlab.http_put(
            url, query_data=query_data, post_data=file_data, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)

        attrs = {
            "package_name": package_name,
            "package_version": package_version,
            "file_name": file_name,
            "path": path,
        }
        attrs.update(server_data)
        return self._obj_cls(self, attrs=attrs)

    @overload
    def download(
        self,
        package_name: str,
        package_version: str,
        file_name: str,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def download(
        self,
        package_name: str,
        package_version: str,
        file_name: str,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def download(
        self,
        package_name: str,
        package_version: str,
        file_name: str,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(
        cls_names="GenericPackageManager",
        required=("package_name", "package_version", "file_name"),
    )
    @exc.on_http_error(exc.GitlabGetError)
    def download(
        self,
        package_name: str,
        package_version: str,
        file_name: str,
        streamed: bool = False,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Download a generic package.

        Args:
            package_name: The package name.
            package_version: The package version.
            file_name: The name of the file in the registry
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server failed to perform the request

        Returns:
            The package content if streamed is False, None otherwise
        """
        path = f"{self._computed_path}/{package_name}/{package_version}/{file_name}"
        result = self.gitlab.http_get(path, streamed=streamed, raw=True, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )


class GroupPackage(RESTObject):
    pass


class GroupPackageManager(ListMixin[GroupPackage]):
    _path = "/groups/{group_id}/packages"
    _obj_cls = GroupPackage
    _from_parent_attrs = {"group_id": "id"}
    _list_filters = (
        "exclude_subgroups",
        "order_by",
        "sort",
        "package_type",
        "package_name",
        "package_version",
        "include_versionless",
        "status",
    )


class ProjectPackage(ObjectDeleteMixin, RESTObject):
    package_files: ProjectPackageFileManager
    pipelines: ProjectPackagePipelineManager


class ProjectPackageManager(
    ListMixin[ProjectPackage], GetMixin[ProjectPackage], DeleteMixin[ProjectPackage]
):
    _path = "/projects/{project_id}/packages"
    _obj_cls = ProjectPackage
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = (
        "order_by",
        "sort",
        "package_type",
        "package_name",
        "package_version",
        "include_versionless",
        "status",
    )


class ProjectPackageFile(ObjectDeleteMixin, RESTObject):
    pass


class ProjectPackageFileManager(
    DeleteMixin[ProjectPackageFile], ListMixin[ProjectPackageFile]
):
    _path = "/projects/{project_id}/packages/{package_id}/package_files"
    _obj_cls = ProjectPackageFile
    _from_parent_attrs = {"project_id": "project_id", "package_id": "id"}


class ProjectPackagePipeline(RESTObject):
    pass


class ProjectPackagePipelineManager(ListMixin[ProjectPackagePipeline]):
    _path = "/projects/{project_id}/packages/{package_id}/pipelines"
    _obj_cls = ProjectPackagePipeline
    _from_parent_attrs = {"project_id": "project_id", "package_id": "id"}



from gitlab import types
from gitlab.base import RESTObject
from gitlab.mixins import ListMixin

__all__ = ["ProjectIterationManager", "GroupIteration", "GroupIterationManager"]


class GroupIteration(RESTObject):
    _repr_attr = "title"


class GroupIterationManager(ListMixin[GroupIteration]):
    _path = "/groups/{group_id}/iterations"
    _obj_cls = GroupIteration
    _from_parent_attrs = {"group_id": "id"}
    # When using the API, the "in" keyword collides with python's "in" keyword
    # raising a SyntaxError.
    # For this reason, we have to use the query_parameters argument:
    # group.iterations.list(query_parameters={"in": "title"})
    _list_filters = (
        "include_ancestors",
        "include_descendants",
        "in",
        "search",
        "state",
        "updated_after",
        "updated_before",
    )
    _types = {"in": types.ArrayAttribute}


class ProjectIterationManager(ListMixin[GroupIteration]):
    _path = "/projects/{project_id}/iterations"
    _obj_cls = GroupIteration
    _from_parent_attrs = {"project_id": "id"}
    # When using the API, the "in" keyword collides with python's "in" keyword
    # raising a SyntaxError.
    # For this reason, we have to use the query_parameters argument:
    # project.iterations.list(query_parameters={"in": "title"})
    _list_filters = (
        "include_ancestors",
        "include_descendants",
        "in",
        "search",
        "state",
        "updated_after",
        "updated_before",
    )
    _types = {"in": types.ArrayAttribute}



from __future__ import annotations

from typing import Any

from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "GroupCluster",
    "GroupClusterManager",
    "ProjectCluster",
    "ProjectClusterManager",
]


class GroupCluster(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class GroupClusterManager(CRUDMixin[GroupCluster]):
    _path = "/groups/{group_id}/clusters"
    _obj_cls = GroupCluster
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "platform_kubernetes_attributes"),
        optional=("domain", "enabled", "managed", "environment_scope"),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "name",
            "domain",
            "management_project_id",
            "platform_kubernetes_attributes",
            "environment_scope",
        )
    )

    @exc.on_http_error(exc.GitlabStopError)
    def create(self, data: dict[str, Any] | None = None, **kwargs: Any) -> GroupCluster:
        """Create a new object.

        Args:
            data: Parameters to send to the server to create the
                         resource
            **kwargs: Extra options to send to the server (e.g. sudo or
                      'ref_name', 'stage', 'name', 'all')

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server cannot perform the request

        Returns:
            A new instance of the manage object class build with
                the data sent by the server
        """
        path = f"{self.path}/user"
        return super().create(data, path=path, **kwargs)


class ProjectCluster(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectClusterManager(CRUDMixin[ProjectCluster]):
    _path = "/projects/{project_id}/clusters"
    _obj_cls = ProjectCluster
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "platform_kubernetes_attributes"),
        optional=("domain", "enabled", "managed", "environment_scope"),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "name",
            "domain",
            "management_project_id",
            "platform_kubernetes_attributes",
            "environment_scope",
        )
    )

    @exc.on_http_error(exc.GitlabStopError)
    def create(
        self, data: dict[str, Any] | None = None, **kwargs: Any
    ) -> ProjectCluster:
        """Create a new object.

        Args:
            data: Parameters to send to the server to create the
                         resource
            **kwargs: Extra options to send to the server (e.g. sudo or
                      'ref_name', 'stage', 'name', 'all')

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server cannot perform the request

        Returns:
            A new instance of the manage object class build with
                the data sent by the server
        """
        path = f"{self.path}/user"
        return super().create(data, path=path, **kwargs)



"""
GitLab API:
https://docs.gitlab.com/ee/api/job_artifacts.html
"""

from __future__ import annotations

from typing import Any, Callable, Iterator, Literal, overload, TYPE_CHECKING

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab import utils
from gitlab.base import RESTManager, RESTObject

__all__ = ["ProjectArtifact", "ProjectArtifactManager"]


class ProjectArtifact(RESTObject):
    """Dummy object to manage custom actions on artifacts"""

    _id_attr = "ref_name"


class ProjectArtifactManager(RESTManager[ProjectArtifact]):
    _obj_cls = ProjectArtifact
    _path = "/projects/{project_id}/jobs/artifacts"
    _from_parent_attrs = {"project_id": "id"}

    @exc.on_http_error(exc.GitlabDeleteError)
    def delete(self, **kwargs: Any) -> None:
        """Delete the project's artifacts on the server.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server cannot perform the request
        """
        path = self._compute_path("/projects/{project_id}/artifacts")

        if TYPE_CHECKING:
            assert path is not None
        self.gitlab.http_delete(path, **kwargs)

    @overload
    def download(
        self,
        ref_name: str,
        job: str,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def download(
        self,
        ref_name: str,
        job: str,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def download(
        self,
        ref_name: str,
        job: str,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(
        cls_names="ProjectArtifactManager",
        required=("ref_name", "job"),
        optional=("job_token",),
    )
    @exc.on_http_error(exc.GitlabGetError)
    def download(
        self,
        ref_name: str,
        job: str,
        streamed: bool = False,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Get the job artifacts archive from a specific tag or branch.

        Args:
            ref_name: Branch or tag name in repository. HEAD or SHA references
                are not supported.
            job: The name of the job.
            job_token: Job token for multi-project pipeline triggers.
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The artifacts if `streamed` is False, None otherwise.
        """
        path = f"{self.path}/{ref_name}/download"
        result = self.gitlab.http_get(
            path, job=job, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )

    @overload
    def raw(
        self,
        ref_name: str,
        artifact_path: str,
        job: str,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def raw(
        self,
        ref_name: str,
        artifact_path: str,
        job: str,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def raw(
        self,
        ref_name: str,
        artifact_path: str,
        job: str,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(
        cls_names="ProjectArtifactManager",
        required=("ref_name", "artifact_path", "job"),
    )
    @exc.on_http_error(exc.GitlabGetError)
    def raw(
        self,
        ref_name: str,
        artifact_path: str,
        job: str,
        streamed: bool = False,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Download a single artifact file from a specific tag or branch from
        within the job's artifacts archive.

        Args:
            ref_name: Branch or tag name in repository. HEAD or SHA references
                are not supported.
            artifact_path: Path to a file inside the artifacts archive.
            job: The name of the job.
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The artifact if `streamed` is False, None otherwise.
        """
        path = f"{self.path}/{ref_name}/raw/{artifact_path}"
        result = self.gitlab.http_get(
            path, streamed=streamed, raw=True, job=job, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )



from gitlab.base import RESTObject
from gitlab.mixins import CreateMixin, DownloadMixin, GetWithoutIdMixin, RefreshMixin
from gitlab.types import RequiredOptional

__all__ = [
    "GroupExport",
    "GroupExportManager",
    "GroupImport",
    "GroupImportManager",
    "ProjectExport",
    "ProjectExportManager",
    "ProjectImport",
    "ProjectImportManager",
]


class GroupExport(DownloadMixin, RESTObject):
    _id_attr = None


class GroupExportManager(GetWithoutIdMixin[GroupExport], CreateMixin[GroupExport]):
    _path = "/groups/{group_id}/export"
    _obj_cls = GroupExport
    _from_parent_attrs = {"group_id": "id"}


class GroupImport(RESTObject):
    _id_attr = None


class GroupImportManager(GetWithoutIdMixin[GroupImport]):
    _path = "/groups/{group_id}/import"
    _obj_cls = GroupImport
    _from_parent_attrs = {"group_id": "id"}


class ProjectExport(DownloadMixin, RefreshMixin, RESTObject):
    _id_attr = None


class ProjectExportManager(
    GetWithoutIdMixin[ProjectExport], CreateMixin[ProjectExport]
):
    _path = "/projects/{project_id}/export"
    _obj_cls = ProjectExport
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(optional=("description",))


class ProjectImport(RefreshMixin, RESTObject):
    _id_attr = None


class ProjectImportManager(GetWithoutIdMixin[ProjectImport]):
    _path = "/projects/{project_id}/import"
    _obj_cls = ProjectImport
    _from_parent_attrs = {"project_id": "id"}



from gitlab.base import RESTObject
from gitlab.mixins import CreateMixin, ListMixin, SaveMixin, UpdateMethod, UpdateMixin
from gitlab.types import RequiredOptional

__all__ = ["ProjectRegistryProtectionRule", "ProjectRegistryProtectionRuleManager"]


class ProjectRegistryProtectionRule(SaveMixin, RESTObject):
    _repr_attr = "repository_path_pattern"


class ProjectRegistryProtectionRuleManager(
    ListMixin[ProjectRegistryProtectionRule],
    CreateMixin[ProjectRegistryProtectionRule],
    UpdateMixin[ProjectRegistryProtectionRule],
):
    _path = "/projects/{project_id}/registry/protection/rules"
    _obj_cls = ProjectRegistryProtectionRule
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("repository_path_pattern",),
        optional=("minimum_access_level_for_push", "minimum_access_level_for_delete"),
    )
    _update_attrs = RequiredOptional(
        optional=(
            "repository_path_pattern",
            "minimum_access_level_for_push",
            "minimum_access_level_for_delete",
        )
    )
    _update_method = UpdateMethod.PATCH



from __future__ import annotations

from typing import Any

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab.base import RESTManager, RESTObject

__all__ = ["SidekiqManager"]


class SidekiqManager(RESTManager[RESTObject]):
    """Manager for the Sidekiq methods.

    This manager doesn't actually manage objects but provides helper function
    for the sidekiq metrics API.
    """

    _path = "/sidekiq"
    _obj_cls = RESTObject

    @cli.register_custom_action(cls_names="SidekiqManager")
    @exc.on_http_error(exc.GitlabGetError)
    def queue_metrics(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Return the registered queues information.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the information couldn't be retrieved

        Returns:
            Information about the Sidekiq queues
        """
        return self.gitlab.http_get(f"{self.path}/queue_metrics", **kwargs)

    @cli.register_custom_action(cls_names="SidekiqManager")
    @exc.on_http_error(exc.GitlabGetError)
    def process_metrics(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Return the registered sidekiq workers.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the information couldn't be retrieved

        Returns:
            Information about the register Sidekiq worker
        """
        return self.gitlab.http_get(f"{self.path}/process_metrics", **kwargs)

    @cli.register_custom_action(cls_names="SidekiqManager")
    @exc.on_http_error(exc.GitlabGetError)
    def job_stats(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Return statistics about the jobs performed.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the information couldn't be retrieved

        Returns:
            Statistics about the Sidekiq jobs performed
        """
        return self.gitlab.http_get(f"{self.path}/job_stats", **kwargs)

    @cli.register_custom_action(cls_names="SidekiqManager")
    @exc.on_http_error(exc.GitlabGetError)
    def compound_metrics(self, **kwargs: Any) -> dict[str, Any] | requests.Response:
        """Return all available metrics and statistics.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the information couldn't be retrieved

        Returns:
            All available Sidekiq metrics and statistics
        """
        return self.gitlab.http_get(f"{self.path}/compound_metrics", **kwargs)



"""
GitLab API:
https://docs.gitlab.com/ee/api/integrations.html
"""

from typing import List

from gitlab import cli
from gitlab.base import RESTObject
from gitlab.mixins import (
    DeleteMixin,
    GetMixin,
    ListMixin,
    ObjectDeleteMixin,
    SaveMixin,
    UpdateMixin,
)

__all__ = [
    "ProjectIntegration",
    "ProjectIntegrationManager",
    "ProjectService",
    "ProjectServiceManager",
]


class ProjectIntegration(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "slug"


class ProjectIntegrationManager(
    GetMixin[ProjectIntegration],
    UpdateMixin[ProjectIntegration],
    DeleteMixin[ProjectIntegration],
    ListMixin[ProjectIntegration],
):
    _path = "/projects/{project_id}/integrations"
    _from_parent_attrs = {"project_id": "id"}
    _obj_cls = ProjectIntegration

    _service_attrs = {
        "asana": (("api_key",), ("restrict_to_branch", "push_events")),
        "assembla": (("token",), ("subdomain", "push_events")),
        "bamboo": (
            ("bamboo_url", "build_key", "username", "password"),
            ("push_events",),
        ),
        "bugzilla": (
            ("new_issue_url", "issues_url", "project_url"),
            ("description", "title", "push_events"),
        ),
        "buildkite": (
            ("token", "project_url"),
            ("enable_ssl_verification", "push_events"),
        ),
        "campfire": (("token",), ("subdomain", "room", "push_events")),
        "circuit": (
            ("webhook",),
            (
                "notify_only_broken_pipelines",
                "branches_to_be_notified",
                "push_events",
                "issues_events",
                "confidential_issues_events",
                "merge_requests_events",
                "tag_push_events",
                "note_events",
                "confidential_note_events",
                "pipeline_events",
                "wiki_page_events",
            ),
        ),
        "custom-issue-tracker": (
            ("new_issue_url", "issues_url", "project_url"),
            ("description", "title", "push_events"),
        ),
        "drone-ci": (
            ("token", "drone_url"),
            (
                "enable_ssl_verification",
                "push_events",
                "merge_requests_events",
                "tag_push_events",
            ),
        ),
        "emails-on-push": (
            ("recipients",),
            (
                "disable_diffs",
                "send_from_committer_email",
                "push_events",
                "tag_push_events",
                "branches_to_be_notified",
            ),
        ),
        "pipelines-email": (
            ("recipients",),
            (
                "add_pusher",
                "notify_only_broken_builds",
                "branches_to_be_notified",
                "notify_only_default_branch",
                "pipeline_events",
            ),
        ),
        "external-wiki": (("external_wiki_url",), ()),
        "flowdock": (("token",), ("push_events",)),
        "github": (("token", "repository_url"), ("static_context",)),
        "hangouts-chat": (
            ("webhook",),
            (
                "notify_only_broken_pipelines",
                "notify_only_default_branch",
                "branches_to_be_notified",
                "push_events",
                "issues_events",
                "confidential_issues_events",
                "merge_requests_events",
                "tag_push_events",
                "note_events",
                "confidential_note_events",
                "pipeline_events",
                "wiki_page_events",
            ),
        ),
        "hipchat": (
            ("token",),
            (
                "color",
                "notify",
                "room",
                "api_version",
                "server",
                "push_events",
                "issues_events",
                "confidential_issues_events",
                "merge_requests_events",
                "tag_push_events",
                "note_events",
                "confidential_note_events",
                "pipeline_events",
            ),
        ),
        "irker": (
            ("recipients",),
            (
                "default_irc_uri",
                "server_port",
                "server_host",
                "colorize_messages",
                "push_events",
            ),
        ),
        "jira": (
            ("url", "username", "password"),
            (
                "api_url",
                "active",
                "jira_issue_transition_id",
                "commit_events",
                "merge_requests_events",
                "comment_on_event_enabled",
            ),
        ),
        "slack-slash-commands": (("token",), ()),
        "mattermost-slash-commands": (("token",), ("username",)),
        "packagist": (
            ("username", "token"),
            ("server", "push_events", "merge_requests_events", "tag_push_events"),
        ),
        "mattermost": (
            ("webhook",),
            (
                "username",
                "channel",
                "notify_only_broken_pipelines",
                "notify_only_default_branch",
                "branches_to_be_notified",
                "push_events",
                "issues_events",
                "confidential_issues_events",
                "merge_requests_events",
                "tag_push_events",
                "note_events",
                "confidential_note_events",
                "pipeline_events",
                "wiki_page_events",
                "push_channel",
                "issue_channel",
                "confidential_issue_channel",
                "merge_request_channel",
                "note_channel",
                "confidential_note_channel",
                "tag_push_channel",
                "pipeline_channel",
                "wiki_page_channel",
            ),
        ),
        "pivotaltracker": (("token",), ("restrict_to_branch", "push_events")),
        "prometheus": (("api_url",), ()),
        "pushover": (
            ("api_key", "user_key", "priority"),
            ("device", "sound", "push_events"),
        ),
        "redmine": (
            ("new_issue_url", "project_url", "issues_url"),
            ("description", "push_events"),
        ),
        "slack": (
            ("webhook",),
            (
                "username",
                "channel",
                "notify_only_broken_pipelines",
                "notify_only_default_branch",
                "branches_to_be_notified",
                "commit_events",
                "confidential_issue_channel",
                "confidential_issues_events",
                "confidential_note_channel",
                "confidential_note_events",
                "deployment_channel",
                "deployment_events",
                "issue_channel",
                "issues_events",
                "job_events",
                "merge_request_channel",
                "merge_requests_events",
                "note_channel",
                "note_events",
                "pipeline_channel",
                "pipeline_events",
                "push_channel",
                "push_events",
                "tag_push_channel",
                "tag_push_events",
                "wiki_page_channel",
                "wiki_page_events",
            ),
        ),
        "microsoft-teams": (
            ("webhook",),
            (
                "notify_only_broken_pipelines",
                "notify_only_default_branch",
                "branches_to_be_notified",
                "push_events",
                "issues_events",
                "confidential_issues_events",
                "merge_requests_events",
                "tag_push_events",
                "note_events",
                "confidential_note_events",
                "pipeline_events",
                "wiki_page_events",
            ),
        ),
        "teamcity": (
            ("teamcity_url", "build_type", "username", "password"),
            ("push_events",),
        ),
        "jenkins": (("jenkins_url", "project_name"), ("username", "password")),
        "mock-ci": (("mock_service_url",), ()),
        "youtrack": (("issues_url", "project_url"), ("description", "push_events")),
    }

    @cli.register_custom_action(
        cls_names=("ProjectIntegrationManager", "ProjectServiceManager")
    )
    def available(self) -> List[str]:
        """List the services known by python-gitlab.

        Returns:
            The list of service code names.
        """
        return list(self._service_attrs.keys())


class ProjectService(ProjectIntegration):
    pass


class ProjectServiceManager(ProjectIntegrationManager):
    _obj_cls = ProjectService



from __future__ import annotations

from typing import Any, Callable, Iterator, Literal, overload, TYPE_CHECKING

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab import utils
from gitlab.base import RESTObject
from gitlab.mixins import RefreshMixin, RetrieveMixin
from gitlab.types import ArrayAttribute

__all__ = ["ProjectJob", "ProjectJobManager"]


class ProjectJob(RefreshMixin, RESTObject):
    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobCancelError)
    def cancel(self, **kwargs: Any) -> dict[str, Any]:
        """Cancel the job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobCancelError: If the job could not be canceled
        """
        path = f"{self.manager.path}/{self.encoded_id}/cancel"
        result = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobRetryError)
    def retry(self, **kwargs: Any) -> dict[str, Any]:
        """Retry the job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobRetryError: If the job could not be retried
        """
        path = f"{self.manager.path}/{self.encoded_id}/retry"
        result = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobPlayError)
    def play(self, **kwargs: Any) -> None:
        """Trigger a job explicitly.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobPlayError: If the job could not be triggered
        """
        path = f"{self.manager.path}/{self.encoded_id}/play"
        result = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        self._update_attrs(result)

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobEraseError)
    def erase(self, **kwargs: Any) -> None:
        """Erase the job (remove job artifacts and trace).

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobEraseError: If the job could not be erased
        """
        path = f"{self.manager.path}/{self.encoded_id}/erase"
        self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabCreateError)
    def keep_artifacts(self, **kwargs: Any) -> None:
        """Prevent artifacts from being deleted when expiration is set.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the request could not be performed
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts/keep"
        self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabCreateError)
    def delete_artifacts(self, **kwargs: Any) -> None:
        """Delete artifacts of a job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the request could not be performed
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts"
        self.manager.gitlab.http_delete(path, **kwargs)

    @overload
    def artifacts(
        self,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def artifacts(
        self,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def artifacts(
        self,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabGetError)
    def artifacts(
        self,
        streamed: bool = False,
        action: Callable[..., Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Get the job artifacts.

        Args:
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The artifacts if `streamed` is False, None otherwise.
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )

    @overload
    def artifact(
        self,
        path: str,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def artifact(
        self,
        path: str,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def artifact(
        self,
        path: str,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabGetError)
    def artifact(
        self,
        path: str,
        streamed: bool = False,
        action: Callable[..., Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Get a single artifact file from within the job's artifacts archive.

        Args:
            path: Path of the artifact
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The artifacts if `streamed` is False, None otherwise.
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts/{path}"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )

    @overload
    def trace(
        self,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def trace(
        self,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def trace(
        self,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabGetError)
    def trace(
        self,
        streamed: bool = False,
        action: Callable[..., Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Get the job trace.

        Args:
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The trace
        """
        path = f"{self.manager.path}/{self.encoded_id}/trace"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )


class ProjectJobManager(RetrieveMixin[ProjectJob]):
    _path = "/projects/{project_id}/jobs"
    _obj_cls = ProjectJob
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("scope", "order_by", "sort")
    _types = {"scope": ArrayAttribute}



from gitlab.base import RESTObject
from gitlab.mixins import (
    CRUDMixin,
    DeleteMixin,
    GetWithoutIdMixin,
    ListMixin,
    ObjectDeleteMixin,
    RefreshMixin,
    SaveMixin,
    UpdateMethod,
    UpdateMixin,
)
from gitlab.types import RequiredOptional

__all__ = [
    "PagesDomain",
    "PagesDomainManager",
    "ProjectPagesDomain",
    "ProjectPagesDomainManager",
    "ProjectPages",
    "ProjectPagesManager",
]


class PagesDomain(RESTObject):
    _id_attr = "domain"


class PagesDomainManager(ListMixin[PagesDomain]):
    _path = "/pages/domains"
    _obj_cls = PagesDomain


class ProjectPagesDomain(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "domain"


class ProjectPagesDomainManager(CRUDMixin[ProjectPagesDomain]):
    _path = "/projects/{project_id}/pages/domains"
    _obj_cls = ProjectPagesDomain
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("domain",), optional=("certificate", "key")
    )
    _update_attrs = RequiredOptional(optional=("certificate", "key"))


class ProjectPages(ObjectDeleteMixin, RefreshMixin, RESTObject):
    _id_attr = None


class ProjectPagesManager(
    DeleteMixin[ProjectPages],
    UpdateMixin[ProjectPages],
    GetWithoutIdMixin[ProjectPages],
):
    _path = "/projects/{project_id}/pages"
    _obj_cls = ProjectPages
    _from_parent_attrs = {"project_id": "id"}
    _update_attrs = RequiredOptional(
        optional=("pages_unique_domain_enabled", "pages_https_only")
    )
    _update_method: UpdateMethod = UpdateMethod.PATCH



from __future__ import annotations

from typing import Any, Callable, Iterator, Literal, overload, TYPE_CHECKING

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab import utils
from gitlab.base import RESTObject, RESTObjectList
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin, UserAgentDetailMixin
from gitlab.types import RequiredOptional

from .award_emojis import ProjectSnippetAwardEmojiManager  # noqa: F401
from .discussions import ProjectSnippetDiscussionManager  # noqa: F401
from .notes import ProjectSnippetNoteManager  # noqa: F401

__all__ = ["Snippet", "SnippetManager", "ProjectSnippet", "ProjectSnippetManager"]


class Snippet(UserAgentDetailMixin, SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "title"

    @overload
    def content(
        self,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def content(
        self,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def content(
        self,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(cls_names="Snippet")
    @exc.on_http_error(exc.GitlabGetError)
    def content(
        self,
        streamed: bool = False,
        action: Callable[..., Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Return the content of a snippet.

        Args:
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment.
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the content could not be retrieved

        Returns:
            The snippet content
        """
        path = f"/snippets/{self.encoded_id}/raw"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )


class SnippetManager(CRUDMixin[Snippet]):
    _path = "/snippets"
    _obj_cls = Snippet
    _create_attrs = RequiredOptional(
        required=("title",),
        exclusive=("files", "file_name"),
        optional=("description", "content", "visibility"),
    )
    _update_attrs = RequiredOptional(
        optional=("title", "files", "file_name", "content", "visibility", "description")
    )

    @overload
    def list_public(
        self, *, iterator: Literal[False] = False, **kwargs: Any
    ) -> list[Snippet]: ...

    @overload
    def list_public(
        self, *, iterator: Literal[True] = True, **kwargs: Any
    ) -> RESTObjectList[Snippet]: ...

    @overload
    def list_public(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[Snippet] | list[Snippet]: ...

    @cli.register_custom_action(cls_names="SnippetManager")
    def list_public(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[Snippet] | list[Snippet]:
        """List all public snippets.

        Args:
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            iterator: If set to True and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabListError: If the list could not be retrieved

        Returns:
            The list of snippets, or a generator if `iterator` is True
        """
        return self.list(path="/snippets/public", iterator=iterator, **kwargs)

    @overload
    def list_all(
        self, *, iterator: Literal[False] = False, **kwargs: Any
    ) -> list[Snippet]: ...

    @overload
    def list_all(
        self, *, iterator: Literal[True] = True, **kwargs: Any
    ) -> RESTObjectList[Snippet]: ...

    @overload
    def list_all(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[Snippet] | list[Snippet]: ...

    @cli.register_custom_action(cls_names="SnippetManager")
    def list_all(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[Snippet] | list[Snippet]:
        """List all snippets.

        Args:
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            iterator: If set to True and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabListError: If the list could not be retrieved

        Returns:
            A generator for the snippets list
        """
        return self.list(path="/snippets/all", iterator=iterator, **kwargs)

    @overload
    def public(
        self,
        *,
        iterator: Literal[False] = False,
        page: int | None = None,
        **kwargs: Any,
    ) -> list[Snippet]: ...

    @overload
    def public(
        self, *, iterator: Literal[True] = True, **kwargs: Any
    ) -> RESTObjectList[Snippet]: ...

    @overload
    def public(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[Snippet] | list[Snippet]: ...

    def public(
        self, *, iterator: bool = False, **kwargs: Any
    ) -> RESTObjectList[Snippet] | list[Snippet]:
        """List all public snippets.

        Args:
            get_all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            iterator: If set to True and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabListError: If the list could not be retrieved

        Returns:
            The list of snippets, or a generator if `iterator` is True
        """
        utils.warn(
            message=(
                "Gitlab.snippets.public() is deprecated and will be removed in a "
                "future major version. Use Gitlab.snippets.list_public() instead."
            ),
            category=DeprecationWarning,
        )
        return self.list(path="/snippets/public", iterator=iterator, **kwargs)


class ProjectSnippet(UserAgentDetailMixin, SaveMixin, ObjectDeleteMixin, RESTObject):
    _url = "/projects/{project_id}/snippets"
    _repr_attr = "title"

    awardemojis: ProjectSnippetAwardEmojiManager
    discussions: ProjectSnippetDiscussionManager
    notes: ProjectSnippetNoteManager

    @overload
    def content(
        self,
        streamed: Literal[False] = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> bytes: ...

    @overload
    def content(
        self,
        streamed: bool = False,
        action: None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[True] = True,
        **kwargs: Any,
    ) -> Iterator[Any]: ...

    @overload
    def content(
        self,
        streamed: Literal[True] = True,
        action: Callable[[bytes], Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @cli.register_custom_action(cls_names="ProjectSnippet")
    @exc.on_http_error(exc.GitlabGetError)
    def content(
        self,
        streamed: bool = False,
        action: Callable[..., Any] | None = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> bytes | Iterator[Any] | None:
        """Return the content of a snippet.

        Args:
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment.
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the content could not be retrieved

        Returns:
            The snippet content
        """
        path = f"{self.manager.path}/{self.encoded_id}/raw"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )


class ProjectSnippetManager(CRUDMixin[ProjectSnippet]):
    _path = "/projects/{project_id}/snippets"
    _obj_cls = ProjectSnippet
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("title", "visibility"),
        exclusive=("files", "file_name"),
        optional=("description", "content"),
    )
    _update_attrs = RequiredOptional(
        optional=("title", "files", "file_name", "content", "visibility", "description")
    )



from __future__ import annotations

from typing import Any

from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ObjectDeleteMixin,
    PromoteMixin,
    RetrieveMixin,
    SaveMixin,
    SubscribableMixin,
    UpdateMixin,
)
from gitlab.types import RequiredOptional

__all__ = ["GroupLabel", "GroupLabelManager", "ProjectLabel", "ProjectLabelManager"]


class GroupLabel(SubscribableMixin, SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    manager: GroupLabelManager

    # Update without ID, but we need an ID to get from list.
    @exc.on_http_error(exc.GitlabUpdateError)
    def save(self, **kwargs: Any) -> None:
        """Saves the changes made to the object to the server.

        The object is updated to match what the server returns.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct.
            GitlabUpdateError: If the server cannot perform the request.
        """
        updated_data = self._get_updated_data()

        # call the manager
        server_data = self.manager.update(None, updated_data, **kwargs)
        self._update_attrs(server_data)


class GroupLabelManager(
    RetrieveMixin[GroupLabel],
    CreateMixin[GroupLabel],
    UpdateMixin[GroupLabel],
    DeleteMixin[GroupLabel],
):
    _path = "/groups/{group_id}/labels"
    _obj_cls = GroupLabel
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "color"), optional=("description", "priority")
    )
    _update_attrs = RequiredOptional(
        required=("name",), optional=("new_name", "color", "description", "priority")
    )

    # Update without ID.
    # NOTE(jlvillal): Signature doesn't match UpdateMixin.update() so ignore
    # type error
    def update(  # type: ignore[override]
        self, name: str | None, new_data: dict[str, Any] | None = None, **kwargs: Any
    ) -> dict[str, Any]:
        """Update a Label on the server.

        Args:
            name: The name of the label
            **kwargs: Extra options to send to the server (e.g. sudo)
        """
        new_data = new_data or {}
        if name:
            new_data["name"] = name
        return super().update(id=None, new_data=new_data, **kwargs)


class ProjectLabel(
    PromoteMixin, SubscribableMixin, SaveMixin, ObjectDeleteMixin, RESTObject
):
    _id_attr = "name"
    manager: ProjectLabelManager

    # Update without ID, but we need an ID to get from list.
    @exc.on_http_error(exc.GitlabUpdateError)
    def save(self, **kwargs: Any) -> None:
        """Saves the changes made to the object to the server.

        The object is updated to match what the server returns.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct.
            GitlabUpdateError: If the server cannot perform the request.
        """
        updated_data = self._get_updated_data()

        # call the manager
        server_data = self.manager.update(None, updated_data, **kwargs)
        self._update_attrs(server_data)


class ProjectLabelManager(
    RetrieveMixin[ProjectLabel],
    CreateMixin[ProjectLabel],
    UpdateMixin[ProjectLabel],
    DeleteMixin[ProjectLabel],
):
    _path = "/projects/{project_id}/labels"
    _obj_cls = ProjectLabel
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name", "color"), optional=("description", "priority")
    )
    _update_attrs = RequiredOptional(
        required=("name",), optional=("new_name", "color", "description", "priority")
    )

    # Update without ID.
    # NOTE(jlvillal): Signature doesn't match UpdateMixin.update() so ignore
    # type error
    def update(  # type: ignore[override]
        self, name: str | None, new_data: dict[str, Any] | None = None, **kwargs: Any
    ) -> dict[str, Any]:
        """Update a Label on the server.

        Args:
            name: The name of the label
            **kwargs: Extra options to send to the server (e.g. sudo)
        """
        new_data = new_data or {}
        if name:
            new_data["name"] = name
        return super().update(id=None, new_data=new_data, **kwargs)



"""
GitLab API:
https://docs.gitlab.com/ee/api/features.html
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from gitlab import exceptions as exc
from gitlab import utils
from gitlab.base import RESTObject
from gitlab.mixins import DeleteMixin, ListMixin, ObjectDeleteMixin

__all__ = ["Feature", "FeatureManager"]


class Feature(ObjectDeleteMixin, RESTObject):
    _id_attr = "name"


class FeatureManager(ListMixin[Feature], DeleteMixin[Feature]):
    _path = "/features/"
    _obj_cls = Feature

    @exc.on_http_error(exc.GitlabSetError)
    def set(
        self,
        name: str,
        value: bool | int,
        feature_group: str | None = None,
        user: str | None = None,
        group: str | None = None,
        project: str | None = None,
        **kwargs: Any,
    ) -> Feature:
        """Create or update the object.

        Args:
            name: The value to set for the object
            value: The value to set for the object
            feature_group: A feature group name
            user: A GitLab username
            group: A GitLab group
            project: A GitLab project in form group/project
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabSetError: If an error occurred

        Returns:
            The created/updated attribute
        """
        name = utils.EncodedId(name)
        path = f"{self.path}/{name}"
        data = {
            "value": value,
            "feature_group": feature_group,
            "user": user,
            "group": group,
            "project": project,
        }
        data = utils.remove_none_from_dict(data)
        server_data = self.gitlab.http_post(path, post_data=data, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(server_data, dict)
        return self._obj_cls(self, server_data)



"""
GitLab API: https://docs.gitlab.com/api/service_accounts/
"""

from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    ObjectRotateMixin,
    RotateMixin,
    SaveMixin,
    UpdateMethod,
    UpdateMixin,
)
from gitlab.types import ArrayAttribute, RequiredOptional

__all__ = [
    "ServiceAccount",
    "ServiceAccountManager",
    "GroupServiceAccount",
    "GroupServiceAccountManager",
    "GroupServiceAccountAccessToken",
    "GroupServiceAccountAccessTokenManager",
    "ProjectServiceAccount",
    "ProjectServiceAccountManager",
    "ProjectServiceAccountAccessToken",
    "ProjectServiceAccountAccessTokenManager",
]

_SA_ACCOUNT_ATTRS = RequiredOptional(optional=("name", "username", "email"))

_SA_TOKEN_CREATE_ATTRS = RequiredOptional(
    required=("name", "scopes"), optional=("description", "expires_at")
)

_SA_TOKEN_LIST_FILTERS = (
    "created_after",
    "created_before",
    "expires_after",
    "expires_before",
    "last_used_after",
    "last_used_before",
    "revoked",
    "search",
    "sort",
    "state",
)


# ---------------------------------------------------------------------------
# Instance-level service accounts
# ---------------------------------------------------------------------------


class ServiceAccount(SaveMixin, RESTObject):
    pass


class ServiceAccountManager(
    CreateMixin[ServiceAccount], ListMixin[ServiceAccount], UpdateMixin[ServiceAccount]
):
    _path = "/service_accounts"
    _obj_cls = ServiceAccount
    _create_attrs = _SA_ACCOUNT_ATTRS
    _update_attrs = _SA_ACCOUNT_ATTRS
    _update_method = UpdateMethod.PATCH
    _list_filters = ("order_by", "sort")


# ---------------------------------------------------------------------------
# Group-level service accounts
# ---------------------------------------------------------------------------


class GroupServiceAccountAccessToken(ObjectDeleteMixin, ObjectRotateMixin, RESTObject):
    pass


class GroupServiceAccountAccessTokenManager(
    CreateMixin[GroupServiceAccountAccessToken],
    DeleteMixin[GroupServiceAccountAccessToken],
    ListMixin[GroupServiceAccountAccessToken],
    RotateMixin[GroupServiceAccountAccessToken],
):
    _path = "/groups/{group_id}/service_accounts/{user_id}/personal_access_tokens"
    _obj_cls = GroupServiceAccountAccessToken
    _from_parent_attrs = {"group_id": "group_id", "user_id": "id"}
    _create_attrs = _SA_TOKEN_CREATE_ATTRS
    _types = {"scopes": ArrayAttribute}
    _list_filters = _SA_TOKEN_LIST_FILTERS


class GroupServiceAccount(SaveMixin, ObjectDeleteMixin, RESTObject):
    access_tokens: GroupServiceAccountAccessTokenManager


class GroupServiceAccountManager(
    CreateMixin[GroupServiceAccount],
    DeleteMixin[GroupServiceAccount],
    ListMixin[GroupServiceAccount],
    UpdateMixin[GroupServiceAccount],
):
    _path = "/groups/{group_id}/service_accounts"
    _obj_cls = GroupServiceAccount
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = _SA_ACCOUNT_ATTRS
    _update_attrs = _SA_ACCOUNT_ATTRS
    _update_method = UpdateMethod.PATCH
    _list_filters = ("order_by", "sort")


# ---------------------------------------------------------------------------
# Project-level service accounts
# ---------------------------------------------------------------------------


class ProjectServiceAccountAccessToken(
    ObjectDeleteMixin, ObjectRotateMixin, RESTObject
):
    pass


class ProjectServiceAccountAccessTokenManager(
    CreateMixin[ProjectServiceAccountAccessToken],
    DeleteMixin[ProjectServiceAccountAccessToken],
    ListMixin[ProjectServiceAccountAccessToken],
    RotateMixin[ProjectServiceAccountAccessToken],
):
    _path = "/projects/{project_id}/service_accounts/{user_id}/personal_access_tokens"
    _obj_cls = ProjectServiceAccountAccessToken
    _from_parent_attrs = {"project_id": "project_id", "user_id": "id"}
    _create_attrs = _SA_TOKEN_CREATE_ATTRS
    _types = {"scopes": ArrayAttribute}
    _list_filters = _SA_TOKEN_LIST_FILTERS


class ProjectServiceAccount(SaveMixin, ObjectDeleteMixin, RESTObject):
    access_tokens: ProjectServiceAccountAccessTokenManager


class ProjectServiceAccountManager(
    CreateMixin[ProjectServiceAccount],
    DeleteMixin[ProjectServiceAccount],
    ListMixin[ProjectServiceAccount],
    UpdateMixin[ProjectServiceAccount],
):
    _path = "/projects/{project_id}/service_accounts"
    _obj_cls = ProjectServiceAccount
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = _SA_ACCOUNT_ATTRS
    _update_attrs = _SA_ACCOUNT_ATTRS
    _update_method = UpdateMethod.PATCH
    _list_filters = ("order_by", "sort")



"""
GitLab API:
https://docs.gitlab.com/api/feature_flags
"""

from __future__ import annotations

from typing import Any

from gitlab import types, utils
from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = ["ProjectFeatureFlag", "ProjectFeatureFlagManager"]


class ProjectFeatureFlag(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "name"
    manager: ProjectFeatureFlagManager

    def _get_save_url_id(self) -> str | int | None:
        """Get the ID used to construct the API URL for the save operation.

        For renames, this must be the *original* name of the flag. For other
        updates, it is the current name.
        """
        if self._id_attr in self._updated_attrs:
            # If the name is being changed, use the original name for the URL.
            obj_id = self._attrs.get(self._id_attr)
            if isinstance(obj_id, str):
                return utils.EncodedId(obj_id)
            return obj_id
        return self.encoded_id

    def save(self, **kwargs: Any) -> dict[str, Any] | None:
        """Save the changes made to the object to the server.

        This is the standard method to use when updating a feature flag object
        that you have already retrieved.

        It is overridden here to correctly handle renaming. When `name` is
        changed, the API requires the *original* name in the URL, and this
        method provides it.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The new object data (*not* a RESTObject)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        updated_data = self._get_updated_data()
        if not updated_data:
            return None

        obj_id = self._get_save_url_id()
        server_data = self.manager.update(obj_id, updated_data, **kwargs)
        self._update_attrs(server_data)
        return server_data


class ProjectFeatureFlagManager(CRUDMixin[ProjectFeatureFlag]):
    _path = "/projects/{project_id}/feature_flags"
    _obj_cls = ProjectFeatureFlag
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("name",), optional=("version", "description", "active", "strategies")
    )
    _update_attrs = RequiredOptional(
        # new_name is used for renaming via CLI and mapped to 'name' in update()
        optional=("name", "new_name", "description", "active", "strategies")
    )
    _list_filters = ("scope",)
    _types = {"strategies": types.JsonAttribute}

    def update(
        self,
        id: str | int | None = None,
        new_data: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Update a Project Feature Flag.

        This is a lower-level method called by `ProjectFeatureFlag.save()` and
        is also used directly by the CLI.

        The `new_name` parameter is a special case to support renaming via the
        CLI (`--new-name`). It is converted to the `name` parameter that the
        GitLab API expects in the request body.

        Args:
            id: The current name of the feature flag.
            new_data: The dictionary of attributes to update.
            **kwargs: Extra options to send to the server (e.g. sudo)
        """
        # Avoid mutating the caller-provided new_data dict by working on a copy.
        data = dict(new_data or {})
        # When used via CLI, we have 'new_name' to distinguish from the ID 'name'.
        # When used via .save(), the object passes 'name' directly in new_data.
        if "new_name" in data:
            data["name"] = data.pop("new_name")
        return super().update(id, data, **kwargs)



"""
GitLab API:
https://docs.gitlab.com/ee/api/instance_level_ci_variables.html
https://docs.gitlab.com/ee/api/project_level_variables.html
https://docs.gitlab.com/ee/api/group_level_variables.html
"""

from gitlab.base import RESTObject
from gitlab.mixins import (
    CreateMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    SaveMixin,
)
from gitlab.types import RequiredOptional

__all__ = [
    "MemberRole",
    "MemberRoleManager",
    "GroupMemberRole",
    "GroupMemberRoleManager",
]


class MemberRole(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class MemberRoleManager(
    ListMixin[MemberRole], CreateMixin[MemberRole], DeleteMixin[MemberRole]
):
    _path = "/member_roles"
    _obj_cls = MemberRole
    _create_attrs = RequiredOptional(
        required=("name", "base_access_level"),
        optional=(
            "description",
            "admin_cicd_variables",
            "admin_compliance_framework",
            "admin_group_member",
            "admin_group_member",
            "admin_merge_request",
            "admin_push_rules",
            "admin_terraform_state",
            "admin_vulnerability",
            "admin_web_hook",
            "archive_project",
            "manage_deploy_tokens",
            "manage_group_access_tokens",
            "manage_merge_request_settings",
            "manage_project_access_tokens",
            "manage_security_policy_link",
            "read_code",
            "read_runners",
            "read_dependency",
            "read_vulnerability",
            "remove_group",
            "remove_project",
        ),
    )


class GroupMemberRole(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class GroupMemberRoleManager(
    ListMixin[GroupMemberRole],
    CreateMixin[GroupMemberRole],
    DeleteMixin[GroupMemberRole],
):
    _path = "/groups/{group_id}/member_roles"
    _from_parent_attrs = {"group_id": "id"}
    _obj_cls = GroupMemberRole
    _create_attrs = RequiredOptional(
        required=("name", "base_access_level"),
        optional=(
            "description",
            "admin_cicd_variables",
            "admin_compliance_framework",
            "admin_group_member",
            "admin_group_member",
            "admin_merge_request",
            "admin_push_rules",
            "admin_terraform_state",
            "admin_vulnerability",
            "admin_web_hook",
            "archive_project",
            "manage_deploy_tokens",
            "manage_group_access_tokens",
            "manage_merge_request_settings",
            "manage_project_access_tokens",
            "manage_security_policy_link",
            "read_code",
            "read_runners",
            "read_dependency",
            "read_vulnerability",
            "remove_group",
            "remove_project",
        ),
    )



from gitlab.base import RESTObject
from gitlab.mixins import ListMixin, RetrieveMixin

__all__ = [
    "Event",
    "EventManager",
    "GroupEpicResourceLabelEvent",
    "GroupEpicResourceLabelEventManager",
    "ProjectEvent",
    "ProjectEventManager",
    "ProjectIssueResourceLabelEvent",
    "ProjectIssueResourceLabelEventManager",
    "ProjectIssueResourceMilestoneEvent",
    "ProjectIssueResourceMilestoneEventManager",
    "ProjectIssueResourceStateEvent",
    "ProjectIssueResourceIterationEventManager",
    "ProjectIssueResourceWeightEventManager",
    "ProjectIssueResourceIterationEvent",
    "ProjectIssueResourceWeightEvent",
    "ProjectIssueResourceStateEventManager",
    "ProjectMergeRequestResourceLabelEvent",
    "ProjectMergeRequestResourceLabelEventManager",
    "ProjectMergeRequestResourceMilestoneEvent",
    "ProjectMergeRequestResourceMilestoneEventManager",
    "ProjectMergeRequestResourceStateEvent",
    "ProjectMergeRequestResourceStateEventManager",
    "UserEvent",
    "UserEventManager",
]


class Event(RESTObject):
    _id_attr = None
    _repr_attr = "target_title"


class EventManager(ListMixin[Event]):
    _path = "/events"
    _obj_cls = Event
    _list_filters = ("action", "target_type", "before", "after", "sort", "scope")


class GroupEpicResourceLabelEvent(RESTObject):
    pass


class GroupEpicResourceLabelEventManager(RetrieveMixin[GroupEpicResourceLabelEvent]):
    _path = "/groups/{group_id}/epics/{epic_id}/resource_label_events"
    _obj_cls = GroupEpicResourceLabelEvent
    _from_parent_attrs = {"group_id": "group_id", "epic_id": "id"}


class ProjectEvent(Event):
    pass


class ProjectEventManager(EventManager):
    _path = "/projects/{project_id}/events"
    _obj_cls = ProjectEvent
    _from_parent_attrs = {"project_id": "id"}


class ProjectIssueResourceLabelEvent(RESTObject):
    pass


class ProjectIssueResourceLabelEventManager(
    RetrieveMixin[ProjectIssueResourceLabelEvent]
):
    _path = "/projects/{project_id}/issues/{issue_iid}/resource_label_events"
    _obj_cls = ProjectIssueResourceLabelEvent
    _from_parent_attrs = {"project_id": "project_id", "issue_iid": "iid"}


class ProjectIssueResourceMilestoneEvent(RESTObject):
    pass


class ProjectIssueResourceMilestoneEventManager(
    RetrieveMixin[ProjectIssueResourceMilestoneEvent]
):
    _path = "/projects/{project_id}/issues/{issue_iid}/resource_milestone_events"
    _obj_cls = ProjectIssueResourceMilestoneEvent
    _from_parent_attrs = {"project_id": "project_id", "issue_iid": "iid"}


class ProjectIssueResourceStateEvent(RESTObject):
    pass


class ProjectIssueResourceStateEventManager(
    RetrieveMixin[ProjectIssueResourceStateEvent]
):
    _path = "/projects/{project_id}/issues/{issue_iid}/resource_state_events"
    _obj_cls = ProjectIssueResourceStateEvent
    _from_parent_attrs = {"project_id": "project_id", "issue_iid": "iid"}


class ProjectIssueResourceIterationEvent(RESTObject):
    pass


class ProjectIssueResourceIterationEventManager(
    RetrieveMixin[ProjectIssueResourceIterationEvent]
):
    _path = "/projects/{project_id}/issues/{issue_iid}/resource_iteration_events"
    _obj_cls = ProjectIssueResourceIterationEvent
    _from_parent_attrs = {"project_id": "project_id", "issue_iid": "iid"}


class ProjectIssueResourceWeightEvent(RESTObject):
    pass


class ProjectIssueResourceWeightEventManager(
    RetrieveMixin[ProjectIssueResourceWeightEvent]
):
    _path = "/projects/{project_id}/issues/{issue_iid}/resource_weight_events"
    _obj_cls = ProjectIssueResourceWeightEvent
    _from_parent_attrs = {"project_id": "project_id", "issue_iid": "iid"}


class ProjectMergeRequestResourceLabelEvent(RESTObject):
    pass


class ProjectMergeRequestResourceLabelEventManager(
    RetrieveMixin[ProjectMergeRequestResourceLabelEvent]
):
    _path = "/projects/{project_id}/merge_requests/{mr_iid}/resource_label_events"
    _obj_cls = ProjectMergeRequestResourceLabelEvent
    _from_parent_attrs = {"project_id": "project_id", "mr_iid": "iid"}


class ProjectMergeRequestResourceMilestoneEvent(RESTObject):
    pass


class ProjectMergeRequestResourceMilestoneEventManager(
    RetrieveMixin[ProjectMergeRequestResourceMilestoneEvent]
):
    _path = "/projects/{project_id}/merge_requests/{mr_iid}/resource_milestone_events"
    _obj_cls = ProjectMergeRequestResourceMilestoneEvent
    _from_parent_attrs = {"project_id": "project_id", "mr_iid": "iid"}


class ProjectMergeRequestResourceStateEvent(RESTObject):
    pass


class ProjectMergeRequestResourceStateEventManager(
    RetrieveMixin[ProjectMergeRequestResourceStateEvent]
):
    _path = "/projects/{project_id}/merge_requests/{mr_iid}/resource_state_events"
    _obj_cls = ProjectMergeRequestResourceStateEvent
    _from_parent_attrs = {"project_id": "project_id", "mr_iid": "iid"}


class UserEvent(Event):
    pass


class UserEventManager(EventManager):
    _path = "/users/{user_id}/events"
    _obj_cls = UserEvent
    _from_parent_attrs = {"user_id": "id"}



from typing import Any

import httpx
from gql.transport.httpx import HTTPXAsyncTransport, HTTPXTransport


class GitlabTransport(HTTPXTransport):
    """A gql httpx transport that reuses an existing httpx.Client.
    By default, gql's transports do not have a keep-alive session
    and do not enable providing your own session that's kept open.
    This transport lets us provide and close our session on our own
    and provide additional auth.
    For details, see https://github.com/graphql-python/gql/issues/91.
    """

    def __init__(self, *args: Any, client: httpx.Client, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.client = client

    def connect(self) -> None:
        pass

    def close(self) -> None:
        pass


class GitlabAsyncTransport(HTTPXAsyncTransport):
    """An async gql httpx transport that reuses an existing httpx.AsyncClient.
    By default, gql's transports do not have a keep-alive session
    and do not enable providing your own session that's kept open.
    This transport lets us provide and close our session on our own
    and provide additional auth.
    For details, see https://github.com/graphql-python/gql/issues/91.
    """

    def __init__(self, *args: Any, client: httpx.AsyncClient, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.client = client

    async def connect(self) -> None:
        pass

    async def close(self) -> None:
        pass



from __future__ import annotations

import dataclasses
import json
from typing import Any, TYPE_CHECKING

from gitlab import exceptions


@dataclasses.dataclass(frozen=True)
class RequiredOptional:
    required: tuple[str, ...] = ()
    optional: tuple[str, ...] = ()
    exclusive: tuple[str, ...] = ()

    def validate_attrs(
        self, *, data: dict[str, Any], excludes: list[str] | None = None
    ) -> None:
        if excludes is None:
            excludes = []

        if self.required:
            required = [k for k in self.required if k not in excludes]
            missing = [attr for attr in required if attr not in data]
            if missing:
                raise AttributeError(f"Missing attributes: {', '.join(missing)}")

        if self.exclusive:
            exclusives = [attr for attr in data if attr in self.exclusive]
            if len(exclusives) > 1:
                raise AttributeError(
                    f"Provide only one of these attributes: {', '.join(exclusives)}"
                )
            if not exclusives:
                raise AttributeError(
                    f"Must provide one of these attributes: "
                    f"{', '.join(self.exclusive)}"
                )


class GitlabAttribute:
    # Used in utils._transform_types() to decide if we should call get_for_api()
    # on the attribute when transform_data is False (e.g. for POST/PUT/PATCH).
    #
    # This allows us to force transformation of data even when sending JSON bodies,
    # which is useful for types like CommaSeparatedStringAttribute.
    transform_in_body = False

    def __init__(self, value: Any = None) -> None:
        self._value = value

    def get(self) -> Any:
        return self._value

    def set_from_cli(self, cli_value: Any) -> None:
        self._value = cli_value

    def get_for_api(self, *, key: str) -> tuple[str, Any]:
        return (key, self._value)


class JsonAttribute(GitlabAttribute):
    def set_from_cli(self, cli_value: str) -> None:
        try:
            self._value = json.loads(cli_value)
        except (ValueError, TypeError) as e:
            raise exceptions.GitlabParsingError(
                f"Could not parse JSON data: {e}"
            ) from e


class _ListArrayAttribute(GitlabAttribute):
    """Helper class to support `list` / `array` types."""

    def set_from_cli(self, cli_value: str) -> None:
        if not cli_value.strip():
            self._value = []
        else:
            self._value = [item.strip() for item in cli_value.split(",")]

    def get_for_api(self, *, key: str) -> tuple[str, str]:
        # Do not comma-split single value passed as string
        if isinstance(self._value, str):
            return (key, self._value)

        if TYPE_CHECKING:
            assert isinstance(self._value, list)
        return (key, ",".join([str(x) for x in self._value]))


class ArrayAttribute(_ListArrayAttribute):
    """To support `array` types as documented in
    https://docs.gitlab.com/ee/api/#array"""

    def get_for_api(self, *, key: str) -> tuple[str, Any]:
        if isinstance(self._value, str):
            return (f"{key}[]", self._value)

        if TYPE_CHECKING:
            assert isinstance(self._value, list)
        return (f"{key}[]", self._value)


class CommaSeparatedListAttribute(_ListArrayAttribute):
    """
    For values which are sent to the server as a Comma Separated Values (CSV) string
    in query parameters (GET), but as a list/array in JSON bodies (POST/PUT).
    """


class CommaSeparatedStringAttribute(_ListArrayAttribute):
    """
    For values which are sent to the server as a Comma Separated Values (CSV) string.
    Unlike CommaSeparatedListAttribute, this type ensures the value is converted
    to a string even in JSON bodies (POST/PUT requests).
    """

    # Used in utils._transform_types() to ensure the value is converted to a string
    # via get_for_api() even when transform_data is False (e.g. for POST/PUT/PATCH).
    # This is needed because some APIs require a CSV string instead of a JSON array.
    transform_in_body = True


class LowercaseStringAttribute(GitlabAttribute):
    def get_for_api(self, *, key: str) -> tuple[str, str]:
        return (key, str(self._value).lower())


class FileAttribute(GitlabAttribute):
    @staticmethod
    def get_file_name(attr_name: str | None = None) -> str | None:
        return attr_name


class ImageAttribute(FileAttribute):
    @staticmethod
    def get_file_name(attr_name: str | None = None) -> str:
        return f"{attr_name}.png" if attr_name else "image.png"



from gitlab import exceptions as exc
from gitlab.base import RESTObject
from gitlab.mixins import CRUDMixin, NoUpdateMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "Hook",
    "HookManager",
    "ProjectHook",
    "ProjectHookManager",
    "GroupHook",
    "GroupHookManager",
]


class Hook(ObjectDeleteMixin, RESTObject):
    _url = "/hooks"
    _repr_attr = "url"


class HookManager(NoUpdateMixin[Hook]):
    _path = "/hooks"
    _obj_cls = Hook
    _create_attrs = RequiredOptional(required=("url",))


class ProjectHook(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "url"

    @exc.on_http_error(exc.GitlabHookTestError)
    def test(self, trigger: str) -> None:
        """
        Test a Project Hook

        Args:
            trigger: Type of trigger event to test

        Raises:
            GitlabHookTestError: If the hook test attempt failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/test/{trigger}"
        self.manager.gitlab.http_post(path)


class ProjectHookManager(CRUDMixin[ProjectHook]):
    _path = "/projects/{project_id}/hooks"
    _obj_cls = ProjectHook
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "job_events",
            "pipeline_events",
            "wiki_page_events",
            "enable_ssl_verification",
            "token",
        ),
    )
    _update_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "job_events",
            "pipeline_events",
            "wiki_events",
            "enable_ssl_verification",
            "token",
        ),
    )


class GroupHook(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "url"

    @exc.on_http_error(exc.GitlabHookTestError)
    def test(self, trigger: str) -> None:
        """
        Test a Group Hook

        Args:
            trigger: Type of trigger event to test

        Raises:
            GitlabHookTestError: If the hook test attempt failed
        """
        path = f"{self.manager.path}/{self.encoded_id}/test/{trigger}"
        self.manager.gitlab.http_post(path)


class GroupHookManager(CRUDMixin[GroupHook]):
    _path = "/groups/{group_id}/hooks"
    _obj_cls = GroupHook
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "confidential_note_events",
            "job_events",
            "pipeline_events",
            "wiki_page_events",
            "deployment_events",
            "releases_events",
            "subgroup_events",
            "enable_ssl_verification",
            "token",
        ),
    )
    _update_attrs = RequiredOptional(
        required=("url",),
        optional=(
            "push_events",
            "issues_events",
            "confidential_issues_events",
            "merge_requests_events",
            "tag_push_events",
            "note_events",
            "confidential_note_events",
            "job_events",
            "pipeline_events",
            "wiki_page_events",
            "deployment_events",
            "releases_events",
            "subgroup_events",
            "enable_ssl_verification",
            "token",
        ),
    )



from __future__ import annotations

from gitlab import types
from gitlab.base import RESTObject
from gitlab.mixins import (
    CRUDMixin,
    DeleteMixin,
    ListMixin,
    ObjectDeleteMixin,
    RetrieveMixin,
    SaveMixin,
)
from gitlab.types import RequiredOptional

__all__ = [
    "GroupBillableMember",
    "GroupBillableMemberManager",
    "GroupBillableMemberMembership",
    "GroupBillableMemberMembershipManager",
    "GroupMember",
    "GroupMemberAll",
    "GroupMemberManager",
    "GroupMemberAllManager",
    "ProjectMember",
    "ProjectMemberAll",
    "ProjectMemberManager",
    "ProjectMemberAllManager",
]


class GroupMember(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "username"


class GroupMemberManager(CRUDMixin[GroupMember]):
    _path = "/groups/{group_id}/members"
    _obj_cls = GroupMember
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("access_level",),
        optional=("expires_at", "tasks_to_be_done"),
        exclusive=("username", "user_id"),
    )
    _update_attrs = RequiredOptional(
        required=("access_level",), optional=("expires_at",)
    )
    _types = {
        "user_ids": types.ArrayAttribute,
        "tasks_to_be_done": types.ArrayAttribute,
    }


class GroupBillableMember(ObjectDeleteMixin, RESTObject):
    _repr_attr = "username"

    memberships: GroupBillableMemberMembershipManager


class GroupBillableMemberManager(
    ListMixin[GroupBillableMember], DeleteMixin[GroupBillableMember]
):
    _path = "/groups/{group_id}/billable_members"
    _obj_cls = GroupBillableMember
    _from_parent_attrs = {"group_id": "id"}
    _list_filters = ("search", "sort")


class GroupBillableMemberMembership(RESTObject):
    _id_attr = "user_id"


class GroupBillableMemberMembershipManager(ListMixin[GroupBillableMemberMembership]):
    _path = "/groups/{group_id}/billable_members/{user_id}/memberships"
    _obj_cls = GroupBillableMemberMembership
    _from_parent_attrs = {"group_id": "group_id", "user_id": "id"}


class GroupMemberAll(RESTObject):
    _repr_attr = "username"


class GroupMemberAllManager(RetrieveMixin[GroupMemberAll]):
    _path = "/groups/{group_id}/members/all"
    _obj_cls = GroupMemberAll
    _from_parent_attrs = {"group_id": "id"}


class ProjectMember(SaveMixin, ObjectDeleteMixin, RESTObject):
    _repr_attr = "username"


class ProjectMemberManager(CRUDMixin[ProjectMember]):
    _path = "/projects/{project_id}/members"
    _obj_cls = ProjectMember
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("access_level",),
        optional=("expires_at", "tasks_to_be_done"),
        exclusive=("username", "user_id"),
    )
    _update_attrs = RequiredOptional(
        required=("access_level",), optional=("expires_at",)
    )
    _types = {
        "user_ids": types.ArrayAttribute,
        "tasks_to_be_dones": types.ArrayAttribute,
    }


class ProjectMemberAll(RESTObject):
    _repr_attr = "username"


class ProjectMemberAllManager(RetrieveMixin[ProjectMemberAll]):
    _path = "/projects/{project_id}/members/all"
    _obj_cls = ProjectMemberAll
    _from_parent_attrs = {"project_id": "id"}
